//  SharpLibVirt.cs
//  
//  Description:
//       This is binding for C libvirt (http://libvirt.org) library to C#.
//
//  Version:
//       0.7.4 (it mean that this unit will works only with this version of libvirt)
//
//  Authors:
//       Jaromír Červenka <cervajz@cervajz.com> ; http://www.cervajz.com/
//  
//  Copyright (c) 2009 Jaromír "Cervajz" Červenka
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
// 
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

// TODO Proverit:
// Kdyz ve virsh zavolam vol-create a potom pool-refresh tak to spadne na drzku. Chce to proverit i v 0.7.4

using System;
using System.IO;
using System.Runtime.InteropServices;

namespace SharpLibVirt
{
	// -- TYPES -- //
	[StructLayout (LayoutKind.Sequential)]
	public class virNodeInfo
	{
		[MarshalAs (UnmanagedType.ByValTStr, SizeConst=32)]
		/// <summary>
		/// String indicating the CPU model.
		/// </summary>
		public string model;
		/// <summary>
		/// Memory size in kilobytes
		/// </summary>
		public ulong memory;
		/// <summary>
		/// The number of active CPUs.
		/// </summary>
		public uint cpus;
		/// <summary>
		/// Expected CPU frequency.
		/// </summary>
		public uint mhz;
		/// <summary>
		/// The number of NUMA cell, 1 for uniform mem access.
		/// </summary>
		public uint nodes;
		/// <summary>
		/// Number of CPU socket per node.
		/// </summary>
		public uint sockets;
		/// <summary>
		/// Number of core per socket.
		/// </summary>
		public uint cores;
		/// <summary>
		/// Number of threads per core.
		/// </summary>
		public uint threads;
	}
	
	public enum virDomainState: int
	{
		/// <summary>
		/// No state.
		/// </summary>
		VIR_DOMAIN_NOSTATE = 0,
		/// <summary>
		/// The domain is running.
		/// </summary>
		VIR_DOMAIN_RUNNING = 1,
		/// <summary>
		/// The domain is blocked on resource.
		/// </summary>
		VIR_DOMAIN_BLOCKED = 2,
		/// <summary>
		/// The domain is paused by user.
		/// </summary>
		VIR_DOMAIN_PAUSED = 3,
		/// <summary>
		/// The domain is being shut down.
		/// </summary>
		VIR_DOMAIN_SHUTDOWN = 4,
		/// <summary>
		/// The domain is shut off.
		/// </summary>
		VIR_DOMAIN_SHUTOFF = 5,
		/// <summary>
		/// The domain is crashed.
		/// </summary>
		VIR_DOMAIN_CRASHED = 6
	}
	
	public enum virDomainXMLFlags: int
	{
		/// <summary>
		/// Dump security sensitive information too.
		/// </summary>
		VIR_DOMAIN_XML_SECURE = 1,
		/// <summary>
		/// Dump inactive domain information.
		/// </summary>
		VIR_DOMAIN_XML_INACTIVE	 =	2
	}
	
	[StructLayout (LayoutKind.Sequential)]
	public class virDomainInfo
	{
		[MarshalAs (UnmanagedType.ByValTStr, SizeConst=32)]
		/// <summary>
		/// The running state, one of virDomainState.
		/// </summary>
		public virDomainState state;
		/// <summary>
		/// The maximum memory in KBytes allowed.
		/// </summary>
		public ulong maxMem;
		/// <summary>
		/// The memory in KBytes used by the domain.
		/// </summary>
		public ulong memory;
		/// <summary>
		/// The number of virtual CPUs for the domain.
		/// </summary>
		public ushort nrVirtCpu;
		/// <summary>
		/// The CPU time used in nanoseconds.
		/// </summary>
		public ulong cpuTime;
	}
	
	[StructLayout (LayoutKind.Sequential)]
	public class virDomainInterfaceStatsStruct
	{
		[MarshalAs (UnmanagedType.ByValTStr, SizeConst=32)]
		public long rx_bytes;
		public long rx_packets;
		public long rx_errs;
		public	long rx_drop;
		public long tx_bytes;
		public long tx_packets;
		public long tx_errs;
		public long tx_drop;
	}
	
	[StructLayoutAttribute (LayoutKind.Sequential)]
	public class virDomainBlockStatsStruct
	{
		[MarshalAs (UnmanagedType.ByValTStr, SizeConst=32)]
		/// <summary>
		/// Number of read requests.
		/// </summary>
		public long rd_req;
		/// <summary>
		/// Number of read bytes.
		/// </summary>
		public long rd_bytes;
		/// <summary>
		/// Number of write requests.
		/// </summary>
		public long wr_req;
		/// <summary>
		/// Number of written bytes.
		/// </summary>
		public long wr_bytes;		
		/// <summary>
		/// In Xen this returns the mysterious 'oo_req'.
		/// </summary>
		public long errs;
	}
	
	public enum virStoragePoolBuildFlags 
	{
		/// <summary>
		/// Regular build from scratch.
		/// </summary>
		VIR_STORAGE_POOL_BUILD_NEW = 0,
		/// <summary>
		/// Repair / reinitialize.
		/// </summary>
		VIR_STORAGE_POOL_BUILD_REPAIR = 1,
		/// <summary>
		/// Extend existing pool.
		/// </summary>
		VIR_STORAGE_POOL_BUILD_RESIZE = 2
	}
	
	[StructLayout (LayoutKind.Sequential)]
	public class virStoragePoolInfo
	{
		[MarshalAs (UnmanagedType.ByValTStr, SizeConst=32)]
		/// <summary>
		/// virStoragePoolState flags
		/// </summary>
		public virStoragePoolState	state;
		/// <summary>
		/// Logical size bytes
		/// </summary>
		public ulong	capacity;
		/// <summary>
		/// Current allocation bytes.
		/// </summary>
		public ulong allocation;
		/// <summary>
		/// Remaining free space bytes.
		/// </summary>
		public ulong available;
	}
	
	public enum virStoragePoolState 
	{
		/// <summary>
		/// Not running.
		/// </summary>
		VIR_STORAGE_POOL_INACTIVE	 =	0,
		/// <summary>
		/// Initializing pool, not available.
		/// </summary>
		VIR_STORAGE_POOL_BUILDING	 =	1,
		/// <summary>
		/// Running normally.
		/// </summary>
		VIR_STORAGE_POOL_RUNNING	 =	2,
		/// <summary>
		/// Running degraded.
		/// </summary>
		VIR_STORAGE_POOL_DEGRADED	 =	3,
	}
	
	public enum virStoragePoolDeleteFlags 
	{
		/// <summary>
		/// Delete metadata only (fast).
		/// </summary>
		VIR_STORAGE_POOL_DELETE_NORMAL = 0, 
		/// <summary>
		/// Clear all data to zeros (slow).
		/// </summary>
		VIR_STORAGE_POOL_DELETE_ZEROED = 1
	}
	
	[StructLayout (LayoutKind.Sequential)]
	public class virStorageVolInfo
	{
		[MarshalAs (UnmanagedType.ByValTStr, SizeConst=32)]
		/// <summary>
		/// virStorageVolType flags.
		/// </summary>
		public virStorageVolType type;
		/// <summary>
		/// Logical size bytes.
		/// </summary>
		public ulong capacity;		
		/// <summary>
		/// Current allocation bytes.
		/// </summary>
		public ulong allocation;
	}

	public enum virStorageVolType
	{
		/// <summary>
		/// Regular file based volumes.
		/// </summary>
		VIR_STORAGE_VOL_FILE = 0,
		/// <summary>
		/// Block based volumes.
		/// </summary>
		VIR_STORAGE_VOL_BLOCK = 1 
	}
	
	[StructLayout (LayoutKind.Sequential)]
	public class virError
	{
		[MarshalAs (UnmanagedType.ByValTStr, SizeConst=32)]
		/// <summary>
		/// The error code, a virErrorNumber.
		/// </summary>
		public virErrorNumber code;
		/// <summary>
		/// What part of the library raised this error.
		/// </summary>
		public virErrorDomain domain;
		/// <summary>
		/// Human-readable informative error message.
		/// </summary>
		public string message;
		/// <summary>
		/// How consequent is the error.
		/// </summary>
		public virErrorLevel level;
		/// <summary>
		/// Connection if available, deprecated see note above.
		/// </summary>
		public IntPtr	conn;
		/// <summary>
		/// Domain if available, deprecated see note above.
		/// </summary>
		public IntPtr	dom;
		/// <summary>
		/// Extra string information.
		/// </summary>
		public string	str1;
		/// <summary>
		/// Extra string information.
		/// </summary>
		public string	str2;
		/// <summary>
		/// Extra string information.
		/// </summary>
		public string 	str3;
		/// <summary>
		/// Extra number information.
		/// </summary>
		public int	int1;
		/// <summary>
		/// Extra number information.
		/// </summary>
		public int	int2;
		/// <summary>
		/// Network if available, deprecated see note above.
		/// </summary>		
		public IntPtr net;
	}
	
	public enum virErrorDomain 
	{
		VIR_FROM_NONE =	0,
		VIR_FROM_XEN =	1,	 
		VIR_FROM_XEND =	2,	 
		VIR_FROM_XENSTORE =	3,
		VIR_FROM_SEXPR = 4,	 
		VIR_FROM_XML = 5,	 
		VIR_FROM_DOM = 6,	
		VIR_FROM_RPC = 7,	
		VIR_FROM_PROXY = 8,
		VIR_FROM_CONF =	9,	
		VIR_FROM_QEMU = 10,	
		VIR_FROM_NET = 11,	
		VIR_FROM_TEST = 12,	
		VIR_FROM_REMOTE = 13,
		VIR_FROM_OPENVZ = 14,	
		VIR_FROM_XENXM = 15,	
		VIR_FROM_STATS_LINUX = 16,
		VIR_FROM_LXC = 17,	 
		VIR_FROM_STORAGE = 18,	
		VIR_FROM_NETWORK = 19,
		VIR_FROM_DOMAIN = 20,	 
		VIR_FROM_UML =	21,
		VIR_FROM_NODEDEV =	22,
		VIR_FROM_XEN_INOTIFY =	23,
		VIR_FROM_SECURITY =	24,
		VIR_FROM_VBOX =	25,
		VIR_FROM_INTERFACE = 26,
		VIR_FROM_ONE =	27,	 
		VIR_FROM_ESX = 28,	 
		VIR_FROM_PHYP =	29,	 
		VIR_FROM_SECRET = 30
	}

	public enum virErrorNumber 
	{
		VIR_ERR_OK = 0,
		/// <summary>
		/// Internal error
		/// </summary>
		VIR_ERR_INTERNAL_ERROR = 1,
		/// <summary>
		/// Memory allocation failure
		/// </summary>
		VIR_ERR_NO_MEMORY = 2,	
		/// <summary>
		/// No support for this function
		/// </summary>
		VIR_ERR_NO_SUPPORT = 3,
		/// <summary>
		/// Could not resolve hostname
		/// </summary>
		VIR_ERR_UNKNOWN_HOST = 4,
		/// <summary>
		/// Can't connect to hypervisor
		/// </summary>
		VIR_ERR_NO_CONNECT = 5,
		/// <summary>
		/// Invalid connection object
		/// </summary>
		VIR_ERR_INVALID_CONN = 6,
		/// <summary>
		/// Invalid domain object
		/// </summary>
		VIR_ERR_INVALID_DOMAIN = 7,
		/// <summary>
		/// Invalid function argument
		/// </summary>
		VIR_ERR_INVALID_ARG = 8,
		/// <summary>
		///  A command to hypervisor failed
		/// </summary>
		VIR_ERR_OPERATION_FAILED = 9,
		/// <summary>
		/// A HTTP GET command to failed
		/// </summary>
		VIR_ERR_GET_FAILED = 10,
		/// <summary>
		/// A HTTP POST command to failed
		/// </summary>
		VIR_ERR_POST_FAILED = 11,
		/// <summary>
		/// Unexpected HTTP error code
		/// </summary>
		VIR_ERR_HTTP_ERROR = 12,
		/// <summary>
		/// Failure to serialize an S-Expr
		/// </summary>
		VIR_ERR_SEXPR_SERIAL = 13,
		/// <summary>
		/// Could not open Xen hypervisor control
		/// </summary>
		VIR_ERR_NO_XEN = 14,
		/// <summary>
		/// Failure doing an hypervisor call
		/// </summary>
		VIR_ERR_XEN_CALL = 15,
		/// <summary>
		/// Unknown OS type
		/// </summary>
		VIR_ERR_OS_TYPE = 16,
		/// <summary>
		/// Missing kernel information
		/// </summary>
		VIR_ERR_NO_KERNEL = 17,
		/// <summary>
		/// Missing root device information
		/// </summary>
		VIR_ERR_NO_ROOT = 18,
		/// <summary>
		/// Missing source device information
		/// </summary>
		VIR_ERR_NO_SOURCE = 19,
		/// <summary>
		/// Missing target device information
		/// </summary>
		VIR_ERR_NO_TARGET = 20,
		/// <summary>
		/// Missing domain name information
		/// </summary>
		VIR_ERR_NO_NAME = 21,
		/// <summary>
		/// Missing domain OS information
		/// </summary>
		VIR_ERR_NO_OS = 22,
		/// <summary>
		/// Missing domain devices information
		/// </summary>
		VIR_ERR_NO_DEVICE = 23,
		/// <summary>
		/// Could not open Xen Store control
		/// </summary>
		VIR_ERR_NO_XENSTORE = 24,
		/// <summary>
		/// Too many drivers registered
		/// </summary>
		VIR_ERR_DRIVER_FULL = 25,
		/// <summary>
		/// Not supported by the drivers (DEPRECATED)
		/// </summary>
		VIR_ERR_CALL_FAILED = 26,
		/// <summary>
		/// An XML description is not well formed or broken
		/// </summary>
		VIR_ERR_XML_ERROR = 27,
		/// <summary>
		/// The domain already exist
		/// </summary>
		VIR_ERR_DOM_EXIST = 28,
		/// <summary>
		/// Operation forbidden on read-only connections
		/// </summary>
		VIR_ERR_OPERATION_DENIED = 29,
		/// <summary>
		/// Failed to open a conf file
		/// </summary>
		VIR_ERR_OPEN_FAILED = 30,
		/// <summary>
		/// Failed to read a conf file
		/// </summary>
		VIR_ERR_READ_FAILED = 31,
		/// <summary>
		/// Failed to parse a conf file
		/// </summary>
		VIR_ERR_PARSE_FAILED = 32,
		/// <summary>
		/// Failed to parse the syntax of a conf file
		/// </summary>
		VIR_ERR_CONF_SYNTAX = 33,
		/// <summary>
		/// Failed to write a conf file
		/// </summary>
		VIR_ERR_WRITE_FAILED = 34,
		/// <summary>
		/// Detail of an XML error
		/// </summary>
		VIR_ERR_XML_DETAIL = 35,
		/// <summary>
		/// Invalid network object
		/// </summary>
		VIR_ERR_INVALID_NETWORK = 36,
		/// <summary>
		/// The network already exist
		/// </summary>
		VIR_ERR_NETWORK_EXIST = 37,
		/// <summary>
		/// General system call failure
		/// </summary>
		VIR_ERR_SYSTEM_ERROR = 38,
		/// <summary>
		/// Some sort of RPC error
		/// </summary>
		VIR_ERR_RPC = 39,
		/// <summary>
		/// Error from a GNUTLS call
		/// </summary>
		VIR_ERR_GNUTLS_ERROR = 40,
		/// <summary>
		/// Failed to start network
		/// </summary>
		VIR_WAR_NO_NETWORK = 41,
		/// <summary>
		/// Domain not found or unexpectedly disappeared
		/// </summary>
		VIR_ERR_NO_DOMAIN = 42,
		/// <summary>
		/// Network not found
		/// </summary>
		VIR_ERR_NO_NETWORK = 43,
		/// <summary>
		/// Invalid MAC address
		/// </summary>
		VIR_ERR_INVALID_MAC = 44,
		/// <summary>
		/// Authentication failed
		/// </summary>
		VIR_ERR_AUTH_FAILED = 45,
		/// <summary>
		/// Invalid storage pool object
		/// </summary>
		VIR_ERR_INVALID_STORAGE_POOL = 46,
		/// <summary>
		/// Invalid storage vol object
		/// </summary>
		VIR_ERR_INVALID_STORAGE_VOL = 47,
		/// <summary>
		/// Failed to start storage
		/// </summary>
		VIR_WAR_NO_STORAGE = 48,
		/// <summary>
		/// Storage pool not found
		/// </summary>
		VIR_ERR_NO_STORAGE_POOL = 49,
		/// <summary>
		/// Storage pool not found
		/// </summary>
		VIR_ERR_NO_STORAGE_VOL = 50,
		/// <summary>
		/// Failed to start node driver
		/// </summary>
		VIR_WAR_NO_NODE = 51,
		/// <summary>
		/// Invalid node device object
		/// </summary>
		VIR_ERR_INVALID_NODE_DEVICE = 52,
		/// <summary>
		/// Node device not found
		/// </summary>
		VIR_ERR_NO_NODE_DEVICE = 53,
		/// <summary>
		/// Security model not found
		/// </summary>
		VIR_ERR_NO_SECURITY_MODEL = 54,	
		/// <summary>
		/// Operation is not applicable at this time
		/// </summary>
		VIR_ERR_OPERATION_INVALID = 55,
		/// <summary>
		/// Failed to start interface driver
		/// </summary>
		VIR_WAR_NO_INTERFACE = 56,
		/// <summary>
		/// Interface driver not running
		/// </summary>
		VIR_ERR_NO_INTERFACE = 57,
		/// <summary>
		/// Invalid interface object
		/// </summary>
		VIR_ERR_INVALID_INTERFACE = 58,
		/// <summary>
		/// More than one matching interface found
		/// </summary>
		VIR_ERR_MULTIPLE_INTERFACES = 59,
		/// <summary>
		/// Failed to start secret storage
		/// </summary>
		VIR_WAR_NO_SECRET = 60,
		/// <summary>
		/// Invalid secret
		/// </summary>
		VIR_ERR_INVALID_SECRET = 61,
		/// <summary>
		/// Secret not found
		/// </summary>
		VIR_ERR_NO_SECRET = 62,
		/// <summary>
		/// Unsupported configuration construct
		/// </summary>
		VIR_ERR_CONFIG_UNSUPPORTED = 63,
		/// <summary>
		/// Timeout occurred during operation
		/// </summary>
		VIR_ERR_OPERATION_TIMEOUT = 64,
		/// <summary>
		/// A migration worked, but making the VM persist on the dest host failed
		/// </summary>
		VIR_ERR_MIGRATE_PERSIST_FAILED = 65,
}
	
	public enum virErrorLevel 
	{
		VIR_ERR_NONE = 0,
		/// <summary>
		/// A simple warning.
		/// </summary>
		VIR_ERR_WARNING = 1,
		/// <summary>
		/// An error.
		/// </summary>
		VIR_ERR_ERROR = 2,
	}
	
	/// <summary>
	/// Signature of a function to use when there is an error raised by the library.
	/// </summary>
	public delegate void virErrorFunc(IntPtr userData, virError error);
	
//	public enum virConnectCredentialType 
//	{
//		/// <summary>
//		/// Identity to act as
//		/// </summary>
//		VIR_CRED_USERNAME = 1,
//		/// <summary>
//		/// Identify to authorize as
//		/// </summary>
//		VIR_CRED_AUTHNAME = 2,
//		/// <summary>
//		/// RFC 1766 languages, comma separated
//		/// </summary>
//		VIR_CRED_LANGUAGE = 3,
//		/// <summary>
//		/// client supplies a nonce
//		/// </summary>
//		VIR_CRED_CNONCE = 4,
//		/// <summary>
//		/// Passphrase secret
//		/// </summary>
//		VIR_CRED_PASSPHRASE = 5,
//		/// <summary>
//		/// Challenge response
//		/// </summary>
//		VIR_CRED_ECHOPROMPT = 6,
//		/// <summary>
//		/// Challenge response
//		/// </summary>
//		VIR_CRED_NOECHOPROMPT = 7,
//		/// <summary>
//		/// Authentication realm
//		/// </summary>
//		VIR_CRED_REALM = 8,
//		/// <summary>
//		/// Externally managed credential More may be added - expect the unexpected
//		/// </summary>
//		VIR_CRED_EXTERNAL = 9 
//	}
//	
//	public delegate int virConnectAuthCallback(virConnectCredentialType cred,  uint ncred, IntPtr cbdata);
	
//	[StructLayout (LayoutKind.Sequential)]
//	public class virConnectAuth
//	{
//		[MarshalAs (UnmanagedType.ByValTStr, SizeConst=32)]
//		/// <summary>
//		/// List of supported virConnectCredentialType values
//		/// </summary>
//		public virConnectCredentialType credtype;
//		public uint ncredtype;
//		/// <summary>
//		/// Callback used to collect credentials
//		/// </summary>
//		public virConnectAuthCallback cb;
//		public IntPtr cbdata;
//	}
	
	// --- LibVirt API --- //
	
	public static class libvirt
	{		
		public const string virtLib = "libvirt.so.0";
		// TODO Check in C source!
		public const int VIR_UUID_BUFLEN = 16;
		public const int VIR_UUID_STRING_BUFLEN = 25;
		
		// --- CONNECT --- //
		
//		TODO virConnectAuthCallbackPtr
		
		/// <summary>
		/// This function closes the connection to the Hypervisor. 
		/// This should not be called if further interaction with the Hypervisor are needed especially if there is 
		/// running domain which need further monitoring by the application.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to the hypervisor connection.
		/// </param>
		/// <returns>
		/// A 0 in case of success or -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectClose(IntPtr conn);
		
//		TODO virConnectDomainEventCallback
		
//		TODO virConnectDomainEventDeregister
		
//		TODO virConnectDomainEventRegister
		
//		TODO virConnectDomainXMLFromNative
		
		/// <summary>
		/// Reads a domain XML configuration document, and generates generates a native configuration file describing the domain. 
		/// The format of the native data is hypervisor dependant.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> connection object.
		/// </param>
		/// <param name="nativeFormat">
		/// A configuration format exporting to.
		/// </param>
		/// <param name="domainXml">
		/// The domain configuration to export.
		/// </param>
		/// <param name="flags">
		/// Currently unused, pass 0.
		/// </param>
		/// <returns>
		/// A 0 terminated UTF-8 encoded native config datafile, or NULL in case of error. The caller must free() the returned value.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virConnectDomainXMLToNative(IntPtr conn, string nativeFormat, string domainXml, uint flags);
		
//		TODO virConnectFindStoragePoolSources
		
		/// <summary>
		/// Provides capabilities of the hypervisor / driver.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to the hypervisor connection.
		/// </param>
		/// <returns>
		/// NULL in case of error, or an XML string defining the capabilities. The client must free the returned string after use.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virConnectGetCapabilities(IntPtr conn);
		
		/// <summary>
		/// This returns the system hostname on which the hypervisor is running (the result of the gethostname system call). 
		/// If we are connected to a remote system, then this returns the hostname of the remote system.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to a hypervisor connection.
		/// </param>
		/// <returns>
		/// The hostname which must be freed by the caller, or NULL if there was an error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virConnectGetHostname(IntPtr conn);
		
		/// <summary>
		/// Provides @libVer, which is the version of libvirt used by the daemon running on the @conn host.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="libVer">
		/// A <see cref="System.UInt64"/>returns the libvirt library version used on the connection (OUT).
		/// </param>
		/// <returns>
		/// -1 in case of failure, 0 otherwise, and values for @libVer have the format major * 1,000,000 + minor * 1,000 + release.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectGetLibVersion(IntPtr conn, out ulong libVer);
		
		/// <summary>
		/// Provides the maximum number of virtual CPUs supported for a guest VM of a specific type. 
		/// The 'type' parameter here corresponds to the 'type' attribute in the <domain> element of the XML.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to the hypervisor connection.
		/// </param>
		/// <param name="type">
		/// Value of the 'type' attribute in the <domain> element.
		/// </param>
		/// <returns>
		/// The maximum of virtual CPU or -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectGetMaxVcpus(IntPtr conn, string type);
		
		/// <summary>
		/// Get the name of the Hypervisor software used.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to the hypervisor connection.
		/// </param>
		/// <returns>
		/// NULL in case of error, a static zero terminated string otherwise. 
		/// See also: http://www.redhat.com/archives/libvir-list/2007-February/msg00096.html
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virConnectGetType(IntPtr conn);
			
		/// <summary>
		/// This returns the URI (name) of the hypervisor connection. 
		/// Normally this is the same as or similar to the string passed to the virConnectOpen/virConnectOpenReadOnly call, 
		/// but the driver may make the URI canonical. If name == NULL was passed to virConnectOpen, 
		/// then the driver will return a non-NULL URI which can be used to connect to the same hypervisor later.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to a hypervisor connection.
		/// </param>
		/// <returns>
		/// The URI string which must be freed by the caller, or NULL if there was an error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virConnectGetURI(IntPtr conn);
		
		/// <summary>
		/// Get the version level of the Hypervisor running. This may work only with hypervisor call, 
		/// i.e. with privileged access to the hypervisor, not with a Read-Only connection.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to the hypervisor connection.
		/// </param>
		/// <param name="hvVer">
		/// A <see cref="System.UInt64"/> return value for the version of the running hypervisor (OUT).
		/// </param>
		/// <returns>
		/// -1 in case of error, 0 otherwise. If the version can't be extracted by lack of capacities returns 0 and @hvVer is 0, 
		/// otherwise @hvVer value is major * 1,000,000 + minor * 1,000 + release.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectGetVersion(IntPtr conn, [Out] out ulong hvVer);
		
		/// <summary>
		/// Determine if the connection to the hypervisor is encrypted.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the connection object.
		/// </param>
		/// <returns>
		/// 1 if encrypted, 0 if not encrypted, -1 on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectIsEncrypted(IntPtr conn);
		
		/// <summary>
		/// Determine if the connection to the hypervisor is secure.
		/// A connection will be classed as secure if it is either encrypted, 
		/// or running over a channel which is not exposed to eavesdropping (eg a UNIX domain socket, or pipe)
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the connection object.
		/// </param>
		/// <returns>
		/// 1 if secure, 0 if secure, -1 on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectIsSecure(IntPtr conn);
		
		/// <summary>
		/// List the defined but inactive domains, stores the pointers to the names in @names.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to the hypervisor connection.
		/// </param>
		/// <param name="names">
		/// A <see cref="IntPtr"/> pointer to an array to store the names.
		/// </param>
		/// <param name="maxnames">
		/// A size of the array.
		/// </param>
		/// <returns>
		/// The number of names provided in the array or -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectListDefinedDomains(IntPtr conn, IntPtr names, int maxnames);
		
//		TODO virConnectListDefinedInterfaces
				
//		TODO virConnectListDefinedNetworks
		
		/// <summary>
		/// Provides the list of names of inactive storage pools upto maxnames. If there are more than maxnames, 
		/// the remaining names will be silently ignored.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to hypervisor connection.
		/// </param>
		/// <param name="names">
		/// A <see cref="IntPtr"/>array of char * to fill with pool names (allocated by caller).
		/// </param>
		/// <param name="maxnames">
		/// A <see cref="System.Int32"/>size of the names array
		/// </param>
		/// <returns>
		/// 0 on success, -1 on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectListDefinedStoragePools(IntPtr conn, IntPtr names, int maxnames);
		
		/// <summary>
		/// Collect the list of active domains, and store their ID in @maxids.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="ids">
		/// A <see cref="System.Int32[]"/>array to collect the list of IDs of active domains.
		/// </param>
		/// <param name="maxids">
		/// A <see cref="System.Int32"/>size of @ids.
		/// </param>
		/// <returns>
		/// A <see cref="System.Int32"/>the number of domain found or -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectListDomains(IntPtr conn, [Out] int[] ids, int maxids);
		
//		TODO virConnectListInterfaces
		
//		TODO virConnectListNetworks
		
//		TODO virConnectListSecrets
		
		/// <summary>
		/// Provides the list of names of active storage pools upto maxnames. If there are more than maxnames, 
		/// the remaining names will be silently ignored.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to hypervisor connection.
		/// </param>
		/// <param name="names">
		/// A <see cref="IntPtr"/>array of char * to fill with pool names (allocated by caller).
		/// </param>
		/// <param name="maxnames">
		/// A <see cref="System.Int32"/>size of the names array.
		/// </param>
		/// <returns>
		/// 0 on success, -1 on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectListStoragePools(IntPtr conn, IntPtr names, int maxnames);
		
		/// <summary>
		/// Provides the number of defined but inactive domains.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to the hypervisor connection.
		/// </param>
		/// <returns>
		/// The number of domain found or -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectNumOfDefinedDomains(IntPtr conn);
		
//		TODO virConnectNumOfDefinedInterfaces
		
//		TODO virConnectNumOfDefinedNetworks
		
		/// <summary>
		/// Provides the number of inactive storage pools.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to hypervisor connection.
		/// </param>
		/// <returns>
		/// The number of pools found, or -1 on error
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectNumOfDefinedStoragePools(IntPtr conn);
		
		/// <summary>
		/// Provides the number of active domains.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to the hypervisor connection.
		/// </param>
		/// <returns>
		/// The number of domain found or -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectNumOfDomains(IntPtr conn);
		
//		TODO virConnectNumOfInterfaces
		
//		TODO virConnectNumOfNetworks
		
//		TODO virConnectNumOfSecrets
		
		/// <summary>
		/// Provides the number of active storage pools.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to hypervisor connection.
		/// </param>
		/// <returns>
		/// The number of pools found, or -1 on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virConnectNumOfStoragePools(IntPtr conn);
		
		/// <summary>
		/// This function should be called first to get a connection to the Hypervisor and xen store. 
		/// If @name is NULL then probing will be done to determine a suitable default driver to activate. 
		/// This involves trying each hypervisor in turn until one successfully opens. 
		/// If the LIBVIRT_DEFAULT_URI environment variable is set, 
		/// then it will be used in preference to probing for a driver. 
		/// If connecting to an unprivileged hypervisor driver which requires the libvirtd daemon to be active, 
		/// it will automatically be launched if not already running. 
		/// This can be prevented by setting the environment variable LIBVIRT_AUTOSTART=0 URIs 
		/// are documented at http://libvirt.org/uri.html
		/// </summary>
		/// <param name="name">
		/// A <see cref="System.String"/>URI of the hypervisor.
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/>a pointer to the hypervisor connection or NULL in case of error
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virConnectOpen([In] string name);
		
//		[DllImport (virtLib)]
//		TODO public static extern IntPtr virConnectOpenAuth(string name, virConnectAuth auth, int flags);
		
		/// <summary>
		/// This function should be called first to get a restricted connection to the library functionalities. 
		/// The set of APIs usable are then restricted on the available methods to control the domains. 
		/// See virConnectOpen for notes about environment variables which can have an effect on opening drivers.
		/// </summary>
		/// <param name="name">
		/// URI of the hypervisor.
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/> pointer to the hypervisor connection or NULL in case of error URIs are documented at http://libvirt.org/uri.html
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virConnectOpenReadOnly([In] string name);
		
		/// <summary>
		/// Increment the reference count on the connection. For each additional call to this method, 
		/// there shall be a corresponding call to virConnectClose to release the reference count, 
		/// once the caller no longer needs the reference to this object. 
		/// This method is typically useful for applications where multiple threads are using a connection, 
		/// and it is required that the connection remain open until all threads have finished using it. ie, 
		/// each new thread using a connection would increment the reference count.
		/// </summary>
		/// <param name="virConnectPtr">
		/// A <see cref="IntPtr"/>the connection to hold a reference on.
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/>0 in case of success, -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virConnectRef(IntPtr virConnectPtr);
		
		// --- DOMAIN --- //
		
		/// <summary>
		/// Create a virtual device attachment to backend.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>pointer to domain object.
		/// </param>
		/// <param name="xml">
		/// Pointer to XML description of one device.
		/// </param>
		/// <returns>
		/// 0 in case of success, -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainAttachDevice(IntPtr domain, string xml);
		
//		TODO virDomainBlockPeek
		
		/// <summary>
		/// This function returns block device (disk) stats for block devices attached to the domain. 
		/// The path parameter is the name of the block device. Get this by calling virDomainGetXMLDesc 
		/// and finding the <target dev='...'> attribute within //domain/devices/disk. 
		/// (For example, "xvda"). Domains may have more than one block device. 
		/// To get stats for each you should make multiple calls to this function. 
		/// Individual fields within the stats structure may be returned as -1, 
		/// which indicates that the hypervisor does not support that particular statistic.
		/// </summary>
		/// <param name="dom">
		/// A <see cref="IntPtr"/>pointer to the domain object.
		/// </param>
		/// <param name="path">
		/// Path to the block device
		/// </param>
		/// <param name="stats">
		/// Block device stats (returned).
		/// </param>
		/// <param name="size">
		/// Size of stats structure.
		/// </param>
		/// <returns>
		/// 0 in case of success or -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainBlockStats(IntPtr dom, string path, virDomainBlockStatsStruct stats, int size);
		
		/// <summary>
		/// This method will dump the core of a domain on a given file for analysis. 
		/// Note that for remote Xen Daemon the file path will be interpreted in the remote host.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/> domain object.
		/// </param>
		/// <param name="to">
		/// Path for the core file.
		/// </param>
		/// <param name="flags">
		/// Extra flags, currently unused.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int	virDomainCoreDump(IntPtr domain, string to, int flags);
		
		/// <summary>
		/// Launch a defined domain. If the call succeed the domain moves from the defined to the running domains pools.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/> pointer to a defined domain.
		/// </param>
		/// <returns>
		/// 0 in case of success, -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainCreate(IntPtr domain);		
		
		/// <summary>
		/// Launch a new guest domain, based on an XML description similar to the one returned by virDomainGetXMLDesc().
		/// This function may requires privileged access to the hypervisor. The domain is not persistent, 
		/// so its definition will disappear when it is destroyed, or if the host is restarted (see virDomainDefineXML() to define persistent domains).
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to the hypervisor connection.
		/// </param>
		/// <param name="xmlDesc">
		/// A string containing an XML description of the domain.
		/// </param>
		/// <param name="flags">
		/// Callers should always pass 0.
		/// </param>
		/// <returns>
		/// A new domain object or NULL in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virDomainCreateXML(IntPtr conn, string xmlDesc, uint flags);
		
		/// <summary>
		/// Define a domain, but does not start it. This definition is persistent, until explicitly undefined with virDomainUndefine(). 
		/// A previous definition for this domain would be overriden if it already exists.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="xml">
		/// The XML description for the domain, preferably in UTF-8.
		/// </param>
		/// <returns>
		/// NULL in case of error, a pointer to the domain otherwise.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virDomainDefineXML(IntPtr conn, string xml);
		
		/// <summary>
		/// Destroy the domain object. The running instance is shutdown if not down already 
		/// and all resources used by it are given back to the hypervisor. 
		/// This does not free the associated virDomainPtr object. This function may require privileged access
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainDestroy(IntPtr domain);
				
		/// <summary>
		/// Destroy a virtual device attachment to backend.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>pointer to domain object.
		/// </param>
		/// <param name="xml">
		/// Pointer to XML description of one device.
		/// </param>
		/// <returns>
		/// 0 in case of success, -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainDetachDevice(IntPtr domain, string xml);
		
		/// <summary>
		/// Free the domain object. The running instance is kept alive. The data structure is freed and should not be used thereafter.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainFree(IntPtr domain);
		
		/// <summary>
		/// Provides a boolean value indicating whether the domain configured to be automatically started when the host machine boots.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <param name="autostart">
		/// The value returned.
		/// </param>
		/// <returns>
		/// -1 in case of error, 0 in case of success.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainGetAutostart(IntPtr domain, out int autostart);
		
		/// <summary>
		/// Provides the connection pointer associated with a domain. 
		/// The reference counter on the connection is not increased by this call. 
		/// WARNING: When writing libvirt bindings in other languages, do not use this function. 
		/// Instead, store the connection and the domain object together.
		/// </summary>
		/// <param name="dom">
		/// A <see cref="IntPtr"/>pointer to a domain.
		/// </param>
		/// <returns>
		/// The <see cref="IntPtr"/>or NULL in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virDomainGetConnect(IntPtr dom);
		
		/// <summary>
		/// Get the hypervisor ID number for the domain.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <returns>
		/// The domain ID number or (unsigned int) -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainGetID(IntPtr domain);
				
		/// <summary>
		/// Extract information about a domain. Note that if the connection used to get the domain is limited only a 
		/// partial set of the information can be extracted.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <param name="info">
		/// Pointer to a virDomainInfo structure allocated by the user.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainGetInfo(IntPtr domain, [Out] virDomainInfo info);
		
		/// <summary>
		/// Retrieve the maximum amount of physical memory allocated to a domain. 
		/// If domain is NULL, then this get the amount of memory reserved to Domain0 i.e. the domain where the application runs.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object or NULL.
		/// </param>
		/// <returns>
		/// the memory size in kilobytes or 0 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern ulong virDomainGetMaxMemory(IntPtr domain);
		
		/// <summary>
		/// Provides the maximum number of virtual CPUs supported for the guest VM. 
		/// If the guest is inactive, this is basically the same as virConnectGetMaxVcpus. 
		/// If the guest is running this will reflect the maximum number of virtual CPUs the guest was booted with.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>pointer to domain object.
		/// </param>
		/// <returns>
		/// The maximum of virtual CPU or -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainGetMaxVcpus(IntPtr domain);		
		
		/// <summary>
		/// Get the public name for that domain.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <returns>
		/// Pointer to the name or NULL, the string need not be deallocated its lifetime will be the same as the domain object.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virDomainGetName(IntPtr domain);
		
		/// <summary>
		/// Get the type of domain operation system.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <returns>
		/// The new string or NULL in case of error, the string must be freed by the caller.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virDomainGetOSType(IntPtr domain);
		
//		TODO virDomainGetSchedulerParameters
		
//		TODO virDomainGetSchedulerType
		
//		TODO virDomainGetSecurityLabel
		
		/// <summary>
		/// Get the UUID for a domain.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>a domain object.
		/// </param>
		/// <param name="uuid">
		/// A <see cref="System.Char[]"/>pointer to a VIR_UUID_BUFLEN bytes array.
		/// </param>
		/// <returns>
		/// A <see cref="System.Int32"/>-1 in case of error, 0 in case of success.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainGetUUID(IntPtr domain, [Out] char[] uuid);
		
		/// <summary>
		/// Get the UUID for a domain as string. For more information about UUID see RFC4122.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <param name="buf">
		/// Pointer to a VIR_UUID_STRING_BUFLEN bytes array.
		/// </param>
		/// <returns>
		/// -1 in case of error, 0 in case of success.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainGetUUIDString(IntPtr domain,[Out] IntPtr buf);
		
//		TODO virDomainGetVcpus
		
		/// <summary>
		/// Provide an XML description of the domain. The description may be reused later to relaunch the domain with virDomainCreateXML().
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>a domain object.
		/// </param>
		/// <param name="flags">
		/// An OR'ed set of virDomainXMLFlags.
		/// </param>
		/// <returns>
		/// A 0 terminated UTF-8 encoded XML instance, or NULL in case of error. the caller must free() the returned value.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virDomainGetXMLDesc(IntPtr domain, virDomainXMLFlags flags);
		
		/// <summary>
		/// This function returns network interface stats for interfaces attached to the domain. 
		/// The path parameter is the name of the network interface. Domains may have more than one network interface. 
		/// To get stats for each you should make multiple calls to this function. 
		/// Individual fields within the stats structure may be returned as -1, 
		/// which indicates that the hypervisor does not support that particular statistic.
		/// </summary>
		/// <param name="dom">
		/// A <see cref="IntPtr"/>pointer to the domain object.
		/// </param>
		/// <param name="path">
		/// Path to the interface.
		/// </param>
		/// <param name="stats">
		/// A <see cref="virDomainInterfaceStatsStruct"/>network interface stats (returned).
		/// </param>
		/// <param name="size">
		/// Size of stats structure.
		/// </param>
		/// <returns>
		/// 0 in case of success or -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)] // TODO Tohle nefunguje stale to hlasi, ze nezname jmeno
		public static extern int virDomainInterfaceStats(IntPtr dom, string path, virDomainInterfaceStatsStruct stats, int size);
		
		/// <summary>
		/// Determine if the domain is currently running.
		/// </summary>
		/// <param name="dom">
		/// A <see cref="IntPtr"/>pointer to the domain object.
		/// </param>
		/// <returns>
		/// 1 if running, 0 if inactive, -1 on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainIsActive(IntPtr dom);
		
		/// <summary>
		/// Determine if the domain has a persistent configuration which means it will still exist after shutting down.
		/// </summary>
		/// <param name="dom">
		/// A <see cref="IntPtr"/>pointer to the domain object.
		/// </param>
		/// <returns>
		/// 1 if persistent, 0 if transient, -1 on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainIsPersistent(IntPtr dom);
		
		/// <summary>
		/// Try to find a domain based on the hypervisor ID number.
		/// Note that this won't work for inactive domains which have an ID of -1, 
		/// in that case a lookup based on the Name or UUId need to be done instead.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="id">
		/// The domain ID number.
		/// </param>
		/// <returns>
		/// A new domain object or NULL in case of failure. If the domain cannot be found, then VIR_ERR_NO_DOMAIN error is raised.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virDomainLookupByID(IntPtr conn, int id);
		
		/// <summary>
		/// Try to lookup a domain on the given hypervisor based on its name.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="name">
		/// A <see cref="System.String"/>name for the domain.
		/// </param>
		/// <returns>
		/// A new domain object or NULL in case of failure. If the domain cannot be found, then VIR_ERR_NO_DOMAIN error is raised.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virDomainLookupByName(IntPtr conn, string name);
		
		/// <summary>
		/// Try to lookup a domain on the given hypervisor based on its UUID.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="uuid">
		/// A <see cref="System.Char[]"/>the raw UUID for the domain.
		/// </param>
		/// <returns>
		/// A new domain object or NULL in case of failure. If the domain cannot be found, then VIR_ERR_NO_DOMAIN error is raised.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virDomainLookupByUUID(IntPtr conn, char[] uuid);
		
		/// <summary>
		/// Try to lookup a domain on the given hypervisor based on its UUID.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="uuidstr">
		/// The string UUID for the domain.
		/// </param>
		/// <returns>
		/// A new domain object or NULL in case of failure. If the domain cannot be found, then VIR_ERR_NO_DOMAIN error is raised.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virDomainLookupByUUIDString(IntPtr conn, string uuidstr);
		
//		TODO virDomainMemoryPeek
		
//		TODO virDomainMigrate
		
//		TODO virDomainMigrateToURI
		
//		TODO virDomainPinVcpu
		
		/// <summary>
		/// Reboot a domain, the domain object is still usable there after but the domain OS is being stopped for a restart. 
		/// Note that the guest OS may ignore the request.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <param name="flags">
		/// Extra flags for the reboot operation, not used yet.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainReboot(IntPtr domain, uint flags);
		
		/// <summary>
		/// Increment the reference count on the domain. For each additional call to this method, 
		/// there shall be a corresponding call to virDomainFree to release the reference count, 
		/// once the caller no longer needs the reference to this object. 
		/// This method is typically useful for applications where multiple threads are using a connection, 
		/// and it is required that the connection remain open until all threads have finished using it. ie, 
		/// each new thread using a domain would increment the reference count.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>the domain to hold a reference on.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainRef(IntPtr domain);
		
		/// <summary>
		/// This method will restore a domain saved to disk by virDomainSave().
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="From">
		/// Path to the file with saved domain.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int	virDomainRestore(IntPtr conn, string From);
		
		/// <summary>
		/// Resume a suspended domain, the process is restarted from the state where it was frozen by calling virSuspendDomain(). 
		/// This function may requires privileged access
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainResume(IntPtr domain);
		
		/// <summary>
		/// This method will suspend a domain and save its memory contents to a file on disk. After the call, if successful, 
		/// the domain is not listed as running anymore (this may be a problem). Use virDomainRestore() to restore a domain after saving.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <param name="to">
		/// Path for the output file.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int	virDomainSave(IntPtr domain, string to);
		
		/// <summary>
		/// Configure the domain to be automatically started when the host machine boots.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <param name="autostart">
		/// Whether the domain should be automatically started 0 or 1.
		/// </param>
		/// <returns>
		/// -1 in case of error, 0 in case of success.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainSetAutostart(IntPtr domain, int autostart);
		
		/// <summary>
		/// Dynamically change the maximum amount of physical memory allocated to a domain. 
		/// If domain is NULL, then this change the amount of memory reserved to Domain0 i.e. 
		/// the domain where the application runs. This function requires privileged access to the hypervisor.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object or NULL.
		/// </param>
		/// <param name="memory">
		/// The memory size in kilobytes.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainSetMaxMemory(IntPtr domain, ulong memory);
		
		/// <summary>
		/// Dynamically change the target amount of physical memory allocated to a domain. 
		/// If domain is NULL, then this change the amount of memory reserved to Domain0 i.e. 
		/// the domain where the application runs. This function may requires privileged access to the hypervisor.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object or NULL.
		/// </param>
		/// <param name="memory">
		/// The memory size in kilobytes.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainSetMemory(IntPtr domain, ulong memory);
		
//		TODO virDomainSetSchedulerParameters
		
		/// <summary>
		/// Dynamically change the number of virtual CPUs used by the domain. 
		/// Note that this call may fail if the underlying virtualization hypervisor does not support it 
		/// or if growing the number is arbitrary limited. This function requires privileged access to the hypervisor.
		/// </summary>
		/// <param name="domain">
		/// Pointer to domain object, or NULL for Domain0
		/// </param>
		/// <param name="nvcpus">
		/// The new number of virtual CPUs for this domain.
		/// </param>
		/// <returns>
		/// 0 in case of success, -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainSetVcpus(IntPtr domain, uint nvcpus);
		
		/// <summary>
		/// Shutdown a domain, the domain object is still usable there after but the domain OS is being stopped. 
		/// Note that the guest OS may ignore the request. 
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainShutdown(IntPtr domain);
		
		/// <summary>
		/// Suspends an active domain, the process is frozen without further access to CPU resources and I/O 
		/// but the memory used by the domain at the hypervisor level will stay allocated. 
		/// Use virDomainResume() to reactivate the domain. This function may requires privileged access.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>domain object.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainSuspend(IntPtr domain);
		
		/// <summary>
		/// Undefine a domain but does not stop it if it is running.
		/// </summary>
		/// <param name="domain">
		/// A <see cref="IntPtr"/>pointer to a defined domain.
		/// </param>
		/// <returns>
		/// 0 in case of success, -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virDomainUndefine(IntPtr domain);
		
		// --- EVENT --- //
		
//		TODO virEventAddHandleFunc
		
//		TODO virEventAddTimeoutFunc
		
//		TODO virEventHandleCallback
		
//		TODO virEventRegisterImpl
		
//		TODO virEventRemoveHandleFunc
		
//		TODO virEventRemoveTimeoutFunc
		
//		TODO virEventRemoveTimeoutFunc
		
//		TODO virEventUpdateHandleFunc
		
//		TODO virEventUpdateTimeoutFunc
		
//		TODO virFreeCallback
		
		// --- LIBRARY --- //
		
		/// <summary>
		/// Provides two information back, @libVer is the version of the library while @typeVer will be the version of the hypervisor 
		/// type @type against which the library was compiled. If @type is NULL, "Xen" is assumed, 
		/// if @type is unknown or not available, an error code will be returned and @typeVer will be 0.
		/// </summary>
		/// <param name="libVer">
		/// A <see cref="System.UInt64"/>return value for the library version (OUT).
		/// </param>
		/// <param name="type">
		/// A <see cref="System.String"/>the type of connection/driver looked at.
		/// </param>
		/// <param name="typeVer">
		/// A <see cref="System.UInt64"/>return value for the version of the hypervisor (OUT).
		/// </param>
		/// <returns>
		/// -1 in case of failure, 0 otherwise, and values for @libVer and @typeVer have the format major * 1,000,000 + minor * 1,000 + release.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virGetVersion ([Out] out ulong libVer, [In] string type, [Out] out ulong typeVer);
		
		/// <summary>
		/// Initialize the library. It's better to call this routine at startup in multithreaded applications to avoid 
		/// potential race when initializing the library.
		/// </summary>
		/// <returns>
		/// 0 in case of success, -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virInitialize();	
				
		// --- INTERFACE --- //
		
//		TODO virInterfaceCreate
		
//		TODO virInterfaceDefineXML
		
//		TODO virInterfaceDestroy
		
//		TODO virInterfaceFree
		
//		TODO virInterfaceGetConnect
		
//		TODO virInterfaceGetMACString
		
//		TODO virInterfaceGetName
		
//		TODO virInterfaceGetXMLDesc
		
//		TODO virInterfaceIsActive
		
//		TODO virInterfaceLookupByMACString
		
//		TODO virInterfaceLookupByName
		
//		TODO virInterfaceRef
		
//		TODO virInterfaceUndefine		
		
		// --- NETWORK --- //
		
//	TODO virNetworkCreate
		
//		TODO virNetworkCreateXML
		
//		TODO virNetworkDefineXML
		
//		TODO virNetworkDestroy
		
//		TODO virNetworkFree
		
//		TODO virNetworkGetAutostart
		
//		TODO virNetworkGetBridgeName
		
//		TODO virNetworkGetConnect
		
//		TODO virNetworkGetName
		
//		TODO virNetworkGetUUID
		
//		TODO virNetworkGetUUIDString
		
//		TODO virNetworkGetXMLDesc
		
//		TODO virNetworkIsActive
		
//		TODO virNetworkIsPersistent
		
//		TODO virNetworkLookupByName
		
//		TODO virNetworkLookupByUUID
		
//		TODO virNetworkLookupByUUIDString
		
		/// <summary>
		/// Increment the reference count on the network. For each additional call to this method, 
		/// there shall be a corresponding call to virNetworkFree to release the reference count, 
		/// once the caller no longer needs the reference to this object. 
		/// This method is typically useful for applications where multiple threads are using a connection, 
		/// and it is required that the connection remain open until all threads have finished using it. ie, 
		/// each new thread using a network would increment the reference count.
		/// </summary>
		/// <param name="network">
		/// A <see cref="IntPtr"/>the network to hold a reference on.
		/// </param>
		/// <returns>
		/// 0 in case of success, -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virNetworkRef(IntPtr network);
		
//		TODO virNetworkSetAutostart
		
//		TODO virNetworkUndefine
		
		// --- NODE --- //
		
//		TODO virNodeDeviceCreateXML
		
//		TODO virNodeDeviceDestroy
		
//		TODO virNodeDeviceDettach
		
//		TODO virNodeDeviceFree
		
//		TODO virNodeDeviceGetName
		
//		TODO virNodeDeviceGetParent
		
//		TODO virNodeDeviceGetXMLDesc
		
//		TODO virNodeDeviceListCaps
		
//		TODO virNodeDeviceLookupByName
		
//		TODO virNodeDeviceNumOfCaps
		
//		TODO virNodeDeviceReAttach
		
//		TODO virNodeDeviceRef
		
//		TODO virNodeDeviceReset
		
//		TODO virNodeGetCellsFreeMemory
		
		/// <summary>
		/// Provides the free memory available on the Node Note: most libvirt APIs provide memory sizes in kilobytes, 
		/// but in this function the returned value is in bytes. Divide by 1024 as necessary.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <returns>
		/// The available free memory in bytes or 0 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern ulong virNodeGetFreeMemory(IntPtr conn);
		
		/// <summary>
		/// Extract hardware information about the node.
		/// </summary>
		/// <param name="h">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="info">
		/// Pointer to a virNodeInfo structure allocated by the user.
		/// </param>
		/// <returns>
		/// 0 in case of success and -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virNodeGetInfo(IntPtr h, [Out] virNodeInfo info);
		
//		TODO virNodeGetSecurityModel
		
		/// <summary>
		/// Collect the list of node devices, and store their names in @names.
		/// If the optional 'cap' argument is non-NULL, then the count will be restricted to devices with the specified capability.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="cap">
		/// Capability name.
		/// </param>
		/// <param name="names">
		/// Array to collect the list of node device names.
		/// </param>
		/// <param name="maxnames">
		/// Size of @names.
		/// </param>
		/// <param name="flags">
		/// Flags (unused, pass 0).
		/// </param>
		/// <returns>
		/// The number of node devices found or -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virNodeListDevices(IntPtr conn, string cap, [Out] IntPtr names, int maxnames, uint flags);
		
		/// <summary>
		/// Provides the number of node devices. If the optional 'cap' argument is non-NULL, 
		/// then the count will be restricted to devices with the specified capability.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="cap">
		/// A <see cref="System.String"/>capability name.
		/// </param>
		/// <param name="flags">
		/// A <see cref="System.UInt32"/>flags (unused, pass 0).
		/// </param>
		/// <returns>
		/// The number of node devices or -1 in case of error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virNodeNumOfDevices(IntPtr conn, string cap, uint flags);
		
		// --- SECRET -- //
		
//		TODO virSecretDefineXML
		
//		TODO virSecretFree
		
//		TODO virSecretGetConnect
		
//		TODO virSecretGetUUID
		
//		TODO virSecretGetUUIDString
		
//		TODO virSecretGetUsageID
		
//		TODO virSecretGetUsageType
		
//		TODO virSecretGetValue
		
//		TODO virSecretGetXMLDesc
		
//		TODO virSecretLookupByUUID
		
//		TODO virSecretLookupByUUIDString
		
//		TODO virSecretLookupByUsage
			
//		TODO virSecretRef
		
//		TODO virSecretSetValue
		
//		TODO virSecretUndefine
		
		// --- STORAGE POOL --- //
		
		/// <summary>
		/// Build the underlying storage pool.
		/// </summary>
		/// <param name="pool">
		/// A pointer to storage pool.
		/// </param>
		/// <param name="flags">
		/// Future flags, use 0 for now.
		/// </param>
		/// <returns>
		/// 0 on success, or -1 upon failure
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolBuild(IntPtr pool, virStoragePoolBuildFlags flags);
		
		/// <summary>
		/// Starts an inactive storage pool.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <param name="flags">
		/// Future flags, use 0 for now.
		/// </param>
		/// <returns>
		/// 0 on success, or -1 if it could not be started.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolCreate(IntPtr pool, uint flags);
		
		/// <summary>
		/// Create a new storage based on its XML description. The pool is not persistent, 
		/// so its definition will disappear when it is destroyed, or if the host is restarted
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to hypervisor connection.
		/// </param>
		/// <param name="xmlDesc">
		/// A <see cref="System.String"/>XML description for new pool.
		/// </param>
		/// <param name="flags">
		/// A <see cref="System.UInt32"/>future flags, use 0 for now.
		/// </param>
		/// <returns>
		/// A virStoragePoolPtr object, or NULL if creation failed.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStoragePoolCreateXML(IntPtr conn, string xmlDesc, uint flags);
		
		/// <summary>
		/// Define a new inactive storage pool based on its XML description. The pool is persistent, 
		/// until explicitly undefined.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to hypervisor connection.
		/// </param>
		/// <param name="xml">
		/// A <see cref="System.String"/>XML description for new pool.
		/// </param>
		/// <param name="flags">
		/// A <see cref="System.UInt32"/>future flags, use 0 for now.
		/// </param>
		/// <returns>
		/// A virStoragePoolPtr object, or NULL if creation failed.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStoragePoolDefineXML(IntPtr conn, string xml, uint flags);
		
		/// <summary>
		/// Delete the underlying pool resources. This is a non-recoverable operation. The virStoragePoolPtr object itself is not free'd.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <param name="flags">
		/// A <see cref="virStoragePoolDeleteFlags"/>flags for obliteration process.
		/// </param>
		/// <returns>
		/// 0 on success, or -1 if it could not be obliterate.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolDelete(IntPtr pool, virStoragePoolDeleteFlags flags);
		
		/// <summary>
		/// Destroy an active storage pool. This will deactivate the pool on the host, 
		/// but keep any persistent config associated with it. 
		/// If it has a persistent config it can later be restarted with virStoragePoolCreate(). 
		/// This does not free the associated virStoragePoolPtr object.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <returns>
		/// 0 on success, or -1 if it could not be destroyed.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolDestroy(IntPtr pool);
		
		/// <summary>
		/// Free a storage pool object, releasing all memory associated with it. Does not change the state of the pool on the host.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <returns>
		/// 0 on success, or -1 if it could not be free'd.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolFree(IntPtr pool);
		
		/// <summary>
		/// Fetches the value of the autostart flag, which determines whether the pool is automatically started at boot time.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <param name="autostart">
		/// A <see cref="System.Int32"/>location in which to store autostart flag.
		/// </param>
		/// <returns>
		/// 0 on success, -1 on failure
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolGetAutostart(IntPtr pool, out int autotart);
		
		/// <summary>
		/// Provides the connection pointer associated with a storage pool. The reference counter on the connection is not increased by this call. 
		/// WARNING: When writing libvirt bindings in other languages, do not use this function. Instead, store the connection and the pool object together.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to a pool.
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/>the virConnectPtr or NULL in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStoragePoolGetConnect(IntPtr pool);
		
		/// <summary>
		/// Get volatile information about the storage pool such as free space / usage summary.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <param name="info">
		/// Pointer at which to store info.
		/// </param>
		/// <returns>
		/// 0 on success, or -1 on failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolGetInfo(IntPtr pool, [Out] virStoragePoolInfo info);
		
		/// <summary>
		/// Fetch the locally unique name of the storage pool.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <returns>
		/// The name of the pool, or NULL on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virStoragePoolGetName(IntPtr pool);
		
		/// <summary>
		/// Fetch the globally unique ID of the storage pool.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>Pointer to storage pool.
		/// </param>
		/// <param name="uuid">
		/// Buffer of VIR_UUID_BUFLEN bytes in size.
		/// </param>
		/// <returns>
		/// 0 on success, or -1 on error
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolGetUUID(IntPtr pool, [Out] char[] uuid);
		
		/// <summary>
		/// Fetch the globally unique ID of the storage pool as a string.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <param name="buf">
		/// A <see cref="IntPtr"/>buffer of VIR_UUID_STRING_BUFLEN bytes in size.
		/// </param>
		/// <returns>
		/// 0 on success, or -1 on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolGetUUIDString(IntPtr pool, [Out] IntPtr buf);
		
		/// <summary>
		/// Fetch an XML document describing all aspects of the storage pool. 
		/// This is suitable for later feeding back into the virStoragePoolCreateXML method.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <param name="flags">
		/// Flags for XML format options (set of virDomainXMLFlags).
		/// </param>
		/// <returns>
		/// A XML document, or NULL on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virStoragePoolGetXMLDesc(IntPtr pool, virDomainXMLFlags flags);
		
		/// <summary>
		/// Determine if the storage pool is currently running.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to the storage pool object.
		/// </param>
		/// <returns>
		/// 1 if running, 0 if inactive, -1 on error.
		/// </returns>
		[DllImport (virtLib)] // TODO Tohle nefunguje! Odpoji to spojeni.
		public static extern int virStoragePoolIsActive(IntPtr pool);
		
		/// <summary>
		/// Determine if the storage pool has a persistent configuration which means it will still exist after shutting down.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to the storage pool object.
		/// </param>
		/// <returns>
		/// 1 if persistent, 0 if transient, -1 on error.
		/// </returns>
		[DllImport (virtLib)] // TODO Tohle nefunguje! Odpoji to spojeni.
		public static extern int virStoragePoolIsPersistent(IntPtr pool);
 		
		/// <summary>
		/// Fetch list of storage volume names, limiting to at most maxnames.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <param name="names">
		/// A <see cref="IntPtr"/>array in which to storage volume names.
		/// </param>
		/// <param name="maxnames">
		/// A <see cref="System.Int32"/>size of names array.
		/// </param>
		/// <returns>
		/// The number of names fetched, or -1 on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolListVolumes(IntPtr pool, IntPtr names, int maxnames);
		
		/// <summary>
		/// Fetch a storage pool based on its unique name.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to hypervisor connection.
		/// </param>
		/// <param name="name">
		/// Name of pool to fetch.
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/>virStoragePoolPtr object, or NULL if no matching pool is found.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStoragePoolLookupByName(IntPtr conn, string name);
		
		/// <summary>
		/// Fetch a storage pool based on its globally unique id.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to hypervisor connection.
		/// </param>
		/// <param name="uuid">
		/// Globally unique id of pool to fetch.
		/// </param>
		/// <returns>
		/// A virStoragePoolPtr object, or NULL if no matching pool is found
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStoragePoolLookupByUUID(IntPtr conn, char[] uuid);
		
		/// <summary>
		/// Fetch a storage pool based on its globally unique id.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to hypervisor connection.
		/// </param>
		/// <param name="uuidstr">
		/// A <see cref="System.String"/>globally unique id of pool to fetch.
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/>object, or NULL if no matching pool is found.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStoragePoolLookupByUUIDString(IntPtr conn, string uuidstr);
		
		/// <summary>
		/// Fetch a storage pool which contains a particular volume.
		/// </summary>
		/// <param name="vol">
		/// A <see cref="IntPtr"/>pointer to storage volume.
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/>object, or NULL if no matching pool is found.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStoragePoolLookupByVolume(IntPtr vol);
		
		/// <summary>
		/// Fetch the number of storage volumes within a pool.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <returns>
		/// A <see cref="System.Int32"/>the number of storage pools, or -1 on failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolNumOfVolumes(IntPtr pool);
		
		/// <summary>
		/// Increment the reference count on the pool. For each additional call to this method, 
		/// there shall be a corresponding call to virStoragePoolFree to release the reference count, 
		/// once the caller no longer needs the reference to this object. 
		/// This method is typically useful for applications where multiple threads are using a connection, 
		/// and it is required that the connection remain open until all threads have finished using it. ie, 
		/// each new thread using a pool would increment the reference count.
		/// </summary>
		/// <param name="pool">
		/// The pool to hold a reference on.
		/// </param>
		/// <returns>
		/// 0 in case of success, -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolRef(IntPtr pool);
		
		/// <summary>
		/// Request that the pool refresh its list of volumes. This may involve communicating with a remote server, 
		/// and/or initializing new devices at the OS layer.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <param name="flags">
		/// A <see cref="System.UInt32"/>flags to control refresh behaviour (currently unused, use 0).
		/// </param>
		/// <returns>
		/// 0 if the volume list was refreshed, -1 on failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolRefresh(IntPtr pool, uint flags);
		
		/// <summary>
		/// Sets the autostart flag.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <param name="autostart">
		/// A <see cref="System.Int32"/>new flag setting.
		/// </param>
		/// <returns>
		/// 0 on success, -1 on failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolSetAutostart(IntPtr pool, int autostart);
		
		/// <summary>
		/// Undefine an inactive storage pool.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <returns>
		/// 0 on success, -1 on failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStoragePoolUndefine(IntPtr pool);
		
		// --- VOLUME -- //
		
		/// <summary>
		/// Create a storage volume within a pool based on an XML description. Not all pools support creation of volumes.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <param name="xmldesc">
		/// A <see cref="System.String"/>description of volume to create.
		/// </param>
		/// <param name="flags">
		/// A <see cref="System.UInt32"/>flags for creation (unused, pass 0).
		/// </param>
		/// <returns>
		/// The storage volume, or NULL on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStorageVolCreateXML(IntPtr pool, string xmldesc, uint flags);
		
		/// <summary>
		/// Create a storage volume in the parent pool, using the 'clonevol' volume as input. 
		/// Information for the new volume (name, perms) are passed via a typical volume XML description.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to parent pool for the new volume.
		/// </param>
		/// <param name="xmldesc">
		/// A <see cref="System.String"/>description of volume to create.
		/// </param>
		/// <param name="clonevol">
		/// A <see cref="IntPtr"/>storage volume to use as input.
		/// </param>
		/// <param name="flags">
		/// A <see cref="System.UInt32"/>flags for creation (unused, pass 0).
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/>the storage volume, or NULL on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStorageVolCreateXMLFrom(IntPtr pool, string xmldesc, IntPtr clonevol, uint flags);
		
		/// <summary>
		/// Delete the storage volume from the pool.
		/// </summary>
		/// <param name="vol">
		/// A <see cref="IntPtr"/>pointer to storage volume.
		/// </param>
		/// <param name="flags">
		/// A <see cref="System.UInt32"/>future flags, use 0 for now.
		/// </param>
		/// <returns>
		/// 0 on success, or -1 on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStorageVolDelete(IntPtr vol, uint flags);
		
		/// <summary>
		/// Release the storage volume handle. The underlying storage volume continues to exist.
		/// </summary>
		/// <param name="vol">
		/// A <see cref="IntPtr"/>pointer to storage volume.
		/// </param>
		/// <returns>
		/// 0 on success, or -1 on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStorageVolFree(IntPtr vol);
		
		/// <summary>
		/// Provides the connection pointer associated with a storage volume. 
		/// The reference counter on the connection is not increased by this call. 
		/// WARNING: When writing libvirt bindings in other languages, do not use this function. 
		/// Instead, store the connection and the volume object together.
		/// </summary>
		/// <param name="vol">
		/// A <see cref="IntPtr"/>
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/>
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStorageVolGetConnect(IntPtr vol);
		
		/// <summary>
		/// Fetches volatile information about the storage volume such as its current allocation.
		/// </summary>
		/// <param name="vol">
		/// A <see cref="IntPtr"/>pointer to storage volume.
		/// </param>
		/// <param name="info">
		/// A <see cref="virStorageVolInfo"/>pointer at which to store info.
		/// </param>
		/// <returns>
		/// 0 on success, or -1 on failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStorageVolGetInfo(IntPtr vol, virStorageVolInfo info);
		
		/// <summary>
		/// Fetch the storage volume key. 
		/// This is globally unique, so the same volume will have the same key no matter what host it is accessed from
		/// </summary>
		/// <param name="vol">
		/// A <see cref="IntPtr"/>pointer to storage volume.
		/// </param>
		/// <returns>
		/// The volume key, or NULL on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virStorageVolGetKey(IntPtr vol);
		
		/// <summary>
		/// Fetch the storage volume name. This is unique within the scope of a pool.
		/// </summary>
		/// <param name="vol">
		/// A <see cref="IntPtr"/>pointer to storage volume.
		/// </param>
		/// <returns>
		/// The volume name, or NULL on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virStorageVolGetName(IntPtr vol);
		
		/// <summary>
		/// Fetch the storage volume path. 
		/// Depending on the pool configuration this is either persistent across hosts, 
		/// or dynamically assigned at pool startup. 
		/// Consult pool documentation for information on getting the persistent naming.
		/// </summary>
		/// <param name="vol">
		/// A <see cref="IntPtr"/>pointer to storage volume.
		/// </param>
		/// <returns>
		/// The storage volume path, or NULL on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virStorageVolGetPath(IntPtr vol);
		
		/// <summary>
		/// Fetch an XML document describing all aspects of the storage volume.
		/// </summary>
		/// <param name="vol">
		/// A <see cref="IntPtr"/>pointer to storage volume.
		/// </param>
		/// <param name="flags">
		/// A <see cref="System.UInt32"/>flags for XML generation (unused, pass 0).
		/// </param>
		/// <returns>
		/// The XML document, or NULL on error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern string virStorageVolGetXMLDesc(IntPtr vol, uint flags);
		
		/// <summary>
		/// Fetch a pointer to a storage volume based on its globally unique key.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to hypervisor connection.
		/// </param>
		/// <param name="key">
		/// A <see cref="System.String"/>globally unique key.
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/>storage volume, or NULL if not found / error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStorageVolLookupByKey(IntPtr conn, string key);		                                            
		
		/// <summary>
		/// Fetch a pointer to a storage volume based on its name within a pool.
		/// </summary>
		/// <param name="pool">
		/// A <see cref="IntPtr"/>pointer to storage pool.
		/// </param>
		/// <param name="name">
		/// A <see cref="System.String"/>name of storage volume.
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/>storage volume, or NULL if not found / error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStorageVolLookupByName(IntPtr pool, string name);
		
		/// <summary>
		/// Fetch a pointer to a storage volume based on its locally (host) unique path.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to hypervisor connection.
		/// </param>
		/// <param name="path">
		/// A <see cref="System.String"/>locally unique path.
		/// </param>
		/// <returns>
		/// A <see cref="IntPtr"/>storage volume, or NULL if not found / error.
		/// </returns>
		[DllImport (virtLib)]
		public static extern IntPtr virStorageVolLookupByPath(IntPtr conn, string path);
		
		/// <summary>
		/// Increment the reference count on the vol. For each additional call to this method, 
		/// there shall be a corresponding call to virStorageVolFree to release the reference count, 
		/// once the caller no longer needs the reference to this object. 
		/// This method is typically useful for applications where multiple threads are using a connection, 
		/// and it is required that the connection remain open until all threads have finished using it. ie, 
		/// each new thread using a vol would increment the reference count.
		/// </summary>
		/// <param name="vol">
		/// A <see cref="IntPtr"/>the vol to hold a reference on.
		/// </param>
		/// <returns>
		/// A <see cref="System.Int32"/>0 in case of success, -1 in case of failure.
		/// </returns>
		[DllImport (virtLib)]
		public static extern int virStorageVolRef(IntPtr vol);
		
		// -- STREAM -- //
		
//		TODO virStreamAbort
		
//		TODO virStreamEventAddCallback
		
//		TODO virStreamEventCallback
		
//		TODO virStreamEventRemoveCallback
		
//		TODO virStreamEventUpdateCallback
		
//		TODO virStreamFinish
		
//		TODO virStreamFree
		
//		TODO virStreamNew
		
//		TODO virStreamRecv
		
//		TODO virStreamRecvAll
		
//		TODO virStreamRef
		
//		TODO virStreamSend
		
//		TODO virStreamSendAll
		
//		TODO virStreamSinkFunc
		
//		TODO virStreamSourceFunc		
	}
	
	// --- virtError API --- //	
	public static class virterror
	{

		// virConnCopyLastError - Deprecated
		// virConnGetLastError - Deprecated
		
		/// <summary>
		/// The error object is kept in thread local storage, so separate threads can safely access this concurrently. 
		/// Reset the last error caught on that connection.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/> pointer to the hypervisor connection.
		/// </param>
		[DllImport (libvirt.virtLib)]
		public static extern void virConnResetLastError(IntPtr conn);
		
		/// <summary>
		/// Set a connection error handling function, if @handler is NULL it will reset to default 
		/// which is to pass error back to the global library handler.
		/// </summary>
		/// <param name="conn">
		/// A <see cref="IntPtr"/>pointer to the hypervisor connection.
		/// </param>
		/// <param name="userData">
		/// A <see cref="IntPtr"/>pointer to the user data provided in the handler callback.
		/// </param>
		/// <param name="handler">
		/// A <see cref="virErrorFunc"/>function to get called in case of error or NULL
		/// </param>
		[DllImport (libvirt.virtLib)]
		public static extern void virConnSetErrorFunc(IntPtr conn, IntPtr userData, virErrorFunc handler);
		
		/// <summary>
		/// Copy the content of the last error caught at the library level. 
		/// The error object is kept in thread local storage, so separate threads can safely access this concurrently. 
		/// One will need to free the result with virResetError().
		/// </summary>
		/// <param name="to">
		/// A <see cref="virError"/> target to receive the copy.
		/// </param>
		/// <returns>
		/// 0 if no error was found and the error code otherwise and -1 in case of parameter error.
		/// </returns>
		[DllImport (libvirt.virtLib)]
		public static extern int virCopyLastError([Out] virError to);
		
		/// <summary>
		/// Default routine reporting an error to stderr.
		/// </summary>
		/// <param name="err">
		/// A <see cref="virError"/> pointer to the error.
		/// </param>
		[DllImport (libvirt.virtLib)]
		public static extern void virDefaultErrorFunc([In] virError err);		
		
		/// <summary>
		/// Resets and frees the given error.
		/// </summary>
		/// <param name="err">
		/// A <see cref="virError"/> error to free.
		/// </param>
		[DllImport (libvirt.virtLib)] 
		public static extern void virFreeError(virError err); // Does not work, anybody know why?
		
		/// <summary>
		/// Provide a pointer to the last error caught at the library level. 
		/// The error object is kept in thread local storage, so separate threads can safely access this concurrently.
		/// </summary>
		/// <returns>
		/// A pointer to the last error or NULL if none occurred.
		/// </returns>
		[DllImport (libvirt.virtLib)]
		public static extern virError virGetLastError();
		
		/// <summary>
		/// Reset the error being pointed to.
		/// </summary>
		/// <param name="err">
		/// A <see cref="virError"/> pointer to the to clean up.
		/// </param>
		[DllImport (libvirt.virtLib)]
		public static extern void virResetError(virError err);
		
		/// <summary>
		/// Reset the last error caught at the library level. The error object is kept in thread local storage, 
		/// so separate threads can safely access this concurrently, only resetting their own error object.
		/// </summary>
		[DllImport (libvirt.virtLib)]
		public static extern void virResetLastError();
		
		/// <summary>
		/// Save the last error into a new error object.
		/// </summary>
		/// <returns>
		/// A <see cref="virError"/> pointer to the copied error or NULL if allocation failed. 
		/// It is the caller's responsibility to free the error with virFreeError().
		/// </returns>
		[DllImport (libvirt.virtLib)]
		public static extern virError virSaveLastError();
		
		/// <summary>
		/// Set a library global error handling function, if @handler is NULL, it will reset to default printing on stderr. 
		/// The error raised there are those for which no handler at the connection level could caught.
		/// </summary>
		/// <param name="userData">
		/// A <see cref="IntPtr"/>pointer to the user data provided in the handler callback.
		/// </param>
		/// <param name="handler">
		/// A <see cref="virErrorFunc"/>function to get called in case of error or NULL.
		/// </param>
		[DllImport (libvirt.virtLib)]
		public static extern void virSetErrorFunc(IntPtr userData, virErrorFunc handler);
		}
	
	/// <summary>
	/// Interface for handling errors from virterror
	/// </summary>
	public class slvError
	{
		virErrorFunc eF = null;
		virError _error = null;
		string _log = string.Empty;
		bool _save = false;
		bool _overwrite = false;
		bool _first = true;
		bool _console = true;
		bool _nSilent = false;
		
		void ErrorFunc(IntPtr userData, virError error)
		{
			_error = error;
			
			string txt = string.Format("{0} (Code: {1} Domain: {2} Level: {3} )",_error.message, error.code, 
			                           error.domain, error.level);
			
			if (_save)
				AddToLogFile(txt);
			
			if ((_console) && (!_nSilent))
				Console.WriteLine(GetFullMsg(txt));
			
			_nSilent = false;
		}
		
		string GetFullMsg(string text)
		{
			string time = DateTime.Now.TimeOfDay.ToString();
			string msg = "[" + time + "]: " + text + "\n";
			return msg;
		}
		
		void AddToLogFile(string text)
		{
			StreamWriter log = new StreamWriter(_log, (!_overwrite), System.Text.Encoding.UTF8);
			log.AutoFlush = true;			
			
			string msg = GetFullMsg(text);
			if ((_first) && (!_overwrite))
			{
				string msg1 = "\n /// ---------------------------------------------------------- /// \n";
				log.Write(msg1);
			}
			
			_first = false;
			
			log.Write(msg);			
			log.Close();
		}
		
		/// <summary>
		/// Next error will not be printed to console (Console.WriteLine). This is only for first next error.
		/// </summary>
		public bool NextSilent
		{
			get {return _nSilent;}			
			set {_nSilent = value;}
		}
		
		/// <summary>
		/// Last error inside virError object
		/// </summary>
		public virError LastError
		{
			get {return _error;}
		}
				
		/// <summary>
		/// Error output to plaintext file
		/// </summary>
		public bool SaveToLogFile
		{
			get {return _save;}
			set	{_save = value;	}
		}
		
		/// <summary>
		/// Error output to console (Console.WriteLine)
		/// </summary>
		public bool OutputToConsole
		{
			get	{return _console;}			
			set	{_console = value;}
		}
		
		/// <summary>
		/// Overwrite LOG file
		/// </summary>
		public bool OverwriteLogFile
		{
			get {return _overwrite;}
			set	{_overwrite = value;}
		}
		
		/// <summary>
		/// Full path to LOG file. Note: Directory must exists.
		/// </summary>
		public string LogFile
		{
			get {return _log;}
			set	
			{
				string d = Path.GetDirectoryName(value);
				if (!Directory.Exists(d))
					throw new DirectoryNotFoundException();
						
				_log = value;
			}
		}
		
		public slvError()
		{
			eF = new virErrorFunc(ErrorFunc);
			if (eF != null)
				virterror.virSetErrorFunc(IntPtr.Zero, eF);
			
			_error = new virError();
		}
		
		~slvError()
		{
			if (eF != null)
				virterror.virSetErrorFunc(IntPtr.Zero, null);
		}	
	}
	
	public struct slvVersion
	{
		/// <summary>
		/// Version of LibVirt library.
		/// </summary>
		public ulong Library;
		/// <summary>
		/// Version of hypervisor software.
		/// </summary>
		public ulong Hypervisor;
		public ulong Type;
		/// <summary>
		/// Version of LibVirt library in human-readable format.
		/// </summary>
		public string LibraryStr;
		/// <summary>
		/// Version of hypervisor in human-readable format.
		/// </summary>
		public string HypervisorStr;
		public string TypeStr;
	}
	
	/// <summary>
	/// Provides interface for connecting to libvirt daemon.
	/// </summary>
	public class slvConnection 
	{
		// Handle of connection - virConnectPtr
		IntPtr _handle = IntPtr.Zero;
		// An error object
		slvError _error = null;
		
		int _maxNames = 50;
				
		// --- CONSTRUCTOR/DESTRUCTOR --- //
		public slvConnection(string URI, bool ReadOnly)
		{
			libvirt.virInitialize();
			_error = new slvError();			
			_handle = IntPtr.Zero;
			
			if (!ReadOnly)
				_handle = libvirt.virConnectOpen(URI.Trim());
			else
				_handle = libvirt.virConnectOpenReadOnly(URI.Trim());
		}
		
		~slvConnection()
		{
			if (_handle != IntPtr.Zero)
				libvirt.virConnectClose(_handle);				
		}
		
		public IntPtr ConnectPtr
		{
			get
			{
				return _handle;
			}
		}		
		
		public slvError Error
		{
			get
			{
				return _error;
			}
		}
		
		// --- CONNECTION --- //
		public bool Connected
		{
			get
			{
				return (_handle != IntPtr.Zero);
			}
		}
		
		/// <summary>
		/// This returns the system hostname on which the hypervisor is running (the result of the gethostname system call). 
		/// If we are connected to a remote system, then this returns the hostname of the remote system.
		/// </summary>
		public string Hostname
		{
			get
			{
				CheckConnected();
				string ret = libvirt.virConnectGetHostname(_handle);
				if (ret == string.Empty)
					throw new Exception("virConnectGetHostname: NULL");
				
				return ret;
			}
		}
		
		/// <summary>
		/// This returns the URI (name) of the hypervisor connection. 
		/// Normally this is the same as or similar to the string passed to the virConnectOpen/virConnectOpenReadOnly call, 
		/// but the driver may make the URI canonical. If name == NULL was passed to virConnectOpen, 
		/// then the driver will return a non-NULL URI which can be used to connect to the same hypervisor later.
		/// </summary>
		public string URI
		{
			get
			{
				CheckConnected();
				string ret = libvirt.virConnectGetURI(_handle);
				if (ret == string.Empty)
					throw new Exception("virConnectGetURI: NULL");
				
				return ret;
			}
		}
		
		/// <summary>
		/// Get the name of the Hypervisor software used.
		/// </summary>
		public string Type
		{
			get
			{
				CheckConnected();
				string ret = libvirt.virConnectGetType(_handle);
				if (ret == string.Empty)
					throw new Exception("virConnectGetType: NULL");
				
				return ret;
			}
		}
		
		public bool IsSecure
		{
			get
			{
				CheckConnected();
				int iRet = libvirt.virConnectIsSecure(_handle);
				if (iRet == -1)
					throw new Exception("virConnectIsSecure: -1");
				
				return (iRet == 1);
			}
		}
		
		public bool IsEncrypted
		{
			get
			{
				CheckConnected();
				int iRet = libvirt.virConnectIsEncrypted(_handle);
				if (iRet == -1)
					throw new Exception("virConnectIsEncrypted: -1");
				
				return (iRet == 1);
			}
		}
		
		// --- Hypervisor --- //
		
		/// <summary>
		/// Provides the maximum number of virtual CPUs supported for a guest VM of a specific type.
		/// </summary>
		/// <param name="type">
		/// The 'type' parameter here corresponds to the 'type' attribute in the <domain> element of the XML.
		/// </param>
		/// <returns>
		/// The maximum of virtual CPU or -1 in case of error.
		/// </returns>
		public int GetMaxVcpus(string type)
		{
			CheckConnected();
			int ret = libvirt.virConnectGetMaxVcpus(_handle, type);
			if (ret == -1)
				throw new Exception("virConnectGetMaxVcpus: -1");
			
			return ret;
		}	
		
		/// <summary>
		/// Provides capabilities of the hypervisor / driver.
		/// </summary>
		public string Capabilities
		{
			get
			{
				CheckConnected();
				string ret = libvirt.virConnectGetCapabilities(_handle);
				if (ret == string.Empty)
					throw new Exception("virConnectGetCapabilities: NULL");
				
				return ret;
			}
		}
		
		/// <summary>
		/// Version of library, hypervisor and driver.
		/// </summary>
		public slvVersion Version
		{
			get
			{
				ulong libVer, hypVer, typeVer;
				int i = libvirt.virGetVersion(out libVer, Type, out typeVer);
				if (i == -1)
					throw new Exception("virGetVersion: -1");
				
				i = libvirt.virConnectGetVersion(_handle, out hypVer);
				if (i == -1)
					throw new Exception("virConnectGetVersion: -1");
				
				slvVersion v;
				v.Library = libVer;
				v.Hypervisor = hypVer;
				v.Type = typeVer;
				
				// Library
				decimal major, minor, release;
				major = Decimal.Ceiling(libVer / 1000000);
				minor = Decimal.Ceiling(libVer / 1000);
				release = libVer - (major * 1000000 + minor * 1000);
				
				v.LibraryStr = string.Format("{0}.{1}.{2}", major, minor, release);
				
				// Hypervisor
				major = Decimal.Ceiling(hypVer / 1000000);
				minor = Decimal.Ceiling(hypVer / 1000);
				release = hypVer - (major * 1000000 + minor * 1000);
				
				v.HypervisorStr = string.Format("{0}.{1}.{2}", major, minor, release);
				 
				// Type
				major = Decimal.Ceiling(typeVer / 1000000);
				minor = Decimal.Ceiling(typeVer / 1000);
				release = typeVer - (major * 1000000 + minor * 1000);
				
				v.TypeStr = string.Format("{0}.{1}.{2}", major, minor, release);
				
				return v;
			}
		}		 
		
		// --- NODE --- //
		
		/// <summary>
		/// Provides the free memory available on the Node. Note: most libvirt APIs provide memory sizes in kilobytes, 
		/// but in this function the returned value is in bytes. Divide by 1024 as necessary.
		/// </summary>
		public ulong NodeFreeMemory
		{
			get
			{
				CheckConnected();
				ulong ret = libvirt.virNodeGetFreeMemory(_handle);
				if (ret == 0)
					throw new Exception("virNodeGetFreeMemory: 0");
				
				return ret;
			}
		}	
		
		/// <summary>
		/// Extract hardware information about the node.
		/// </summary>
		public virNodeInfo NodeInfo
		{
			get 
			{
				CheckConnected();
				virNodeInfo ret = new virNodeInfo();
				int i = libvirt.virNodeGetInfo(_handle, ret);
				if (i == -1)
					throw new Exception("virNodeGetInfo: -1");
				
				return ret;
			}
		}		
		
		public int NodeNumOfDevices(string cap)
		{
			CheckConnected();
			int iRet = libvirt.virNodeNumOfDevices(_handle, cap, 0);
			if (iRet == -1)
				throw new Exception("virNodeNumOfDevices: -1");
			
			return iRet;
		}
		
		/// <summary>
		/// Collect the list of node devices.
		/// </summary>
		/// <param name="max">
		/// Size of array.
		/// </param>
		public string[] NodeListDevices(string cap)
		{
			CheckConnected();
			IntPtr ret = Mono.Unix.UnixMarshal.AllocHeap(_maxNames);
			int iRet = libvirt.virNodeListDevices(_handle, cap, ret, _maxNames, 0);
			if (iRet == -1)
				throw new Exception("virNodeListDevices: -1");
			
			return Mono.Unix.UnixMarshal.PtrToStringArray(iRet, ret);
		}
		
		// --- DOMAINS --- //	
		
		public bool DomainCreateXMLString(string XMLString)
		{
			CheckConnected();
			IntPtr i = libvirt.virDomainCreateXML(_handle, XMLString.Trim(), 0);
			return (i != IntPtr.Zero);
		}
		
		/// <summary>
		/// Launch a new guest domain, based on an XML description similar to the one returned by virDomainGetXMLDesc().
		/// This function may requires privileged access to the hypervisor. 
		/// The domain is not persistent, so its definition will disappear when it is destroyed, or if the host is restarted 
		/// (see virDomainDefineXML() to define persistent domains).
		/// </summary>
		/// <param name="pathToXMLFile">
		/// String containing an XML file with description of the domain.
		/// </param>
		public bool DomainCreateXMLFile(string pathToXMLFile)
		{
			CheckConnected();
			if (!CheckFile(pathToXMLFile))
				return false;
			
			string xml = File.ReadAllText(pathToXMLFile);
			return DomainCreateXMLString(xml);
		}
		
		public bool DomainDefineXMLString(string XMLString)
		{
			CheckConnected();
			IntPtr i = libvirt.virDomainDefineXML(_handle, XMLString.Trim());
			return (i != IntPtr.Zero);
		}
		
		/// <summary>
		/// Define a domain, but does not start it. This definition is persistent, until explicitly undefined with virDomainUndefine(). 
		/// A previous definition for this domain would be overriden if it already exists.
		/// </summary>
		/// <param name="pathToXMLFile">
		/// Path to XML file which contains description for the domain.
		/// </param>
		public bool DomainDefineXMLFile(string pathToXMLFile)
		{
			CheckConnected();
			if (!CheckFile(pathToXMLFile))
				return false;
			
			string xml = File.ReadAllText(pathToXMLFile);
			return DomainDefineXMLString(xml);
		}
		
		/// <summary>
		/// Reads a domain XML configuration document, and generates generates a native configuration file describing the domain. 
		/// The format of the native data is hypervisor dependant.
		/// </summary>
		/// <param name="nativeFormat">
		/// Configuration format exporting to.
		/// </param>
		/// <param name="pathToDomainXMLFile">
		/// The domain configuration to export (path to file).
		/// </param>
		/// <returns>
		/// A 0 terminated UTF-8 encoded native config datafile, or NULL in case of error.
		/// </returns>
		public string DomainXMLToNative(string nativeFormat, string pathToDomainXMLFile)
		{
			CheckConnected();
			if (!CheckFile(pathToDomainXMLFile))
				return string.Empty;
			
			string xml = System.IO.File.ReadAllText(pathToDomainXMLFile);
			return libvirt.virConnectDomainXMLToNative(_handle, nativeFormat.Trim(), xml, 0);
		}
		
		/// <summary>
		/// Try to lookup a domain on the given hypervisor based on its name.
		/// </summary>
		/// <param name="name">
		/// Name for the domain.
		/// </param>
		/// <returns>
		/// A new domain object or NULL in case of failure. If the domain cannot be found, then VIR_ERR_NO_DOMAIN error is raised.
		/// </returns>
		public IntPtr DomainLookupByName(string name)
		{
			CheckConnected();
			return libvirt.virDomainLookupByName(_handle, name.Trim());
		}
		
		public slvDomain DomLookupByName(string name)
		{
			CheckConnected();
			slvDomain ret = new slvDomain(this, name);
			if (ret.Loaded)
				return ret;
			else
				return null;
		}
		
		/// <summary>
		/// Try to find a domain based on the hypervisor ID number. 
		/// Note that this won't work for inactive domains which have an ID of -1, 
		/// in that case a lookup based on the Name or UUId need to be done instead.
		/// </summary>
		/// <param name="id">
		/// The domain ID number.
		/// </param>
		/// <returns>
		/// A new domain object or NULL in case of failure. If the domain cannot be found, then VIR_ERR_NO_DOMAIN error is raised.
		/// </returns>
		public IntPtr DomainLookupByID(int id)
		{
			CheckConnected();
			return libvirt.virDomainLookupByID(_handle, id);
		}		
		
		/// <summary>
		/// Try to lookup a domain on the given hypervisor based on its UUID.
		/// </summary>
		/// <param name="uuidString">
		/// The string UUID for the domain.
		/// </param>
		/// <returns>
		/// A new domain object or NULL in case of failure. If the domain cannot be found, then VIR_ERR_NO_DOMAIN error is raised.
		/// </returns>
		public IntPtr DomainLookupByUUIDString(string uuidString)
		{
			CheckConnected();
			return libvirt.virDomainLookupByUUIDString(_handle, uuidString.Trim());
		}
		
		public IntPtr DomainLookupByUUID(char[] uuid)
		{
			CheckConnected();
			return libvirt.virDomainLookupByUUID(_handle, uuid);
		}
		
		/// <summary>
		/// Check if domain exists.
		/// </summary>
		/// <param name="name">
		/// A name of domain.
		/// </param>
		public bool IsDomainExist(string name)
		{
			CheckConnected();
			_error.NextSilent = true;
			IntPtr i = DomainLookupByName(name);
			return (i != IntPtr.Zero);
		}
		
		/// <summary>
		/// Check if domain exists.
		/// </summary>
		/// <param name="name">
		/// A UUID string of domain.
		/// </param>
		public bool IsDomainExistUUID(string uuidString)
		{
			CheckConnected();
			_error.NextSilent = true;
			IntPtr i = DomainLookupByUUIDString(uuidString);
			return (i != IntPtr.Zero);
		}
		
		public bool IsDomainExistUUID(char[] uuid)
		{
			CheckConnected();
			_error.NextSilent = true;
			IntPtr i = DomainLookupByUUID(uuid);
			return (i != IntPtr.Zero);
		}
		
		/// <summary>
		/// List the defined but inactive domains, stores the names in array.
		/// </summary>
		public string[] ListInactiveDomains
		{
			get
			{
				CheckConnected();
				IntPtr names = Mono.Unix.UnixMarshal.AllocHeap(_maxNames);
				int iRet = libvirt.virConnectListDefinedDomains(_handle, names, _maxNames);
				if (iRet == -1)
					throw new Exception("virConnectListDefinedDomains: -1");
            
				return Mono.Unix.UnixMarshal.PtrToStringArray(iRet, names);
			}
		}		
		
		public string[] ListOfDomains
		{
			get
			{
				string [] tmp = ListInactiveDomains;
				int [] iTmp = ActiveDomains;
				for (int i = 0; i < iTmp.Length; i++)
				{
					IntPtr dom = DomainLookupByID(iTmp[i]);
					if (dom != IntPtr.Zero)
					{
						Array.Resize(ref tmp, tmp.Length + 1);
						tmp[tmp.Length - 1] = libvirt.virDomainGetName(dom);
					}
				}
				
				Array.Sort(tmp);
				return tmp;
			}
		}
		
		/// <summary>
		/// Collect the list of active domains, and store their IDs in array.
		/// </summary>
		public int[] ActiveDomains
		{
			get
			{
				CheckConnected();
				int [] ids = new int[_maxNames];
				int iRet = libvirt.virConnectListDomains(_handle, ids, _maxNames);
				if (iRet == -1)
					throw new Exception("virConnectListDomains: -1");
            
				Array.Resize(ref ids, iRet);
			
				return ids;
			}
		}
		
		/// <summary>
		/// Provides the number of active domains.
		/// </summary>
		public int NumOfActiveDomains 
		{
			get
			{
				CheckConnected();
				int ret = libvirt.virConnectNumOfDomains(_handle);
				if (ret == -1)
					throw new Exception("virConnectNumOfDomains: -1");
				
				return ret;
			}
		}
			
		/// <summary>
		/// Provides the number of defined but inactive domains.
		/// </summary>
		public int NumOfInactiveDomains
		{
			get
			{
				CheckConnected();
				int ret = libvirt.virConnectNumOfDefinedDomains(_handle);
				if (ret == -1)
					throw new Exception("virConnectNumOfDefinedDomains: -1");
				return ret;
			}
		}
		
		// --- STORAGE --- //
		public IntPtr StoragePoolLookupByName(string name)
		{
			CheckConnected();
			return libvirt.virStoragePoolLookupByName(_handle, name.Trim());
		}
		
		public slvStoragePool StorPoolLookupByName(string name)
		{
			CheckConnected();
			slvStoragePool ret = new slvStoragePool(this, name);
			if (ret.Loaded)
				return ret;
			else
				return null;
		}
		
		public IntPtr StoragePoolLookupByUUIDString(string uuidString)
		{
			CheckConnected();
			return (libvirt.virStoragePoolLookupByUUIDString(_handle, uuidString.Trim()));
		}
		
		public IntPtr StoragePoolLookupByUUID(char[] uuid)
		{
			CheckConnected();
			return (libvirt.virStoragePoolLookupByUUID(_handle, uuid));
		}
		
		public IntPtr StorageVolLookupByKey(string key)
		{
			CheckConnected();
			return (libvirt.virStorageVolLookupByKey(_handle, key.Trim()));
		}
		
		public IntPtr virStorageVolLookupByPath(string path)
		{
			CheckConnected();
			return (libvirt.virStorageVolLookupByPath(_handle, path.Trim()));
		}
		
		public bool IsStorageExist(string name)
		{
			CheckConnected();
			_error.NextSilent = true;
			IntPtr ret = StoragePoolLookupByName(name);
			return (ret != IntPtr.Zero);
		}
		
		public bool IsStorageExistUUID(string uuidString)
		{
			CheckConnected();
			_error.NextSilent = true;
			IntPtr ret = StoragePoolLookupByUUIDString(uuidString);
			return (ret != IntPtr.Zero);
		}
		
		public bool IsStorageExistUUID(char[] uuid)
		{
			CheckConnected();
			_error.NextSilent = true;
			IntPtr ret = StoragePoolLookupByUUID(uuid);
			return (ret != IntPtr.Zero);
		}
		
		/// <summary>
		/// Provides the number of active storage pools.
		/// </summary>
		public int NumOfActiveStoragePools
		{
			get
			{
				CheckConnected();
				int iRet = libvirt.virConnectNumOfStoragePools(_handle);
				if (iRet == -1)
					throw new Exception("virConnectNumOfStoragePools: -1");
				
				return iRet;
			}
		}
		
		/// <summary>
		/// Provides the number of inactive storage pools.
		/// </summary>
		public int NumOfInactiveStoragePools
		{
			get
			{
				CheckConnected();
				int iRet = libvirt.virConnectNumOfDefinedStoragePools(_handle);
				if (iRet == -1)
					throw new Exception("virConnectNumOfDefinedStoragePools: -1");
				
				return iRet;
			}
		}
		
		public string[] ListActiveStoragePools
		{
			get
			{
				CheckConnected();
				IntPtr ret = Mono.Unix.UnixMarshal.AllocHeap(_maxNames);
				int iRet = libvirt.virConnectListStoragePools(_handle, ret, _maxNames);
				if (iRet == -1)
					throw new Exception("virConnectListStoragePools: -1");
			
				return Mono.Unix.UnixMarshal.PtrToStringArray(iRet, ret);
			}
		}
		
		public string[] ListInactiveStoragePools
		{
			get
			{
				CheckConnected();
				IntPtr ret = Mono.Unix.UnixMarshal.AllocHeap(_maxNames);
				int iRet = libvirt.virConnectListDefinedStoragePools(_handle, ret, _maxNames);
				if (iRet == -1)
					throw new Exception("virConnectListDefinedStoragePools: -1");
				
				return Mono.Unix.UnixMarshal.PtrToStringArray(iRet, ret);
			}
		}
		
		public bool StoragePoolDefineXMLString(string XMLString)
		{
			CheckConnected();
			IntPtr i = libvirt.virStoragePoolDefineXML(_handle, XMLString.Trim(), 0);
			return (i != IntPtr.Zero);
		}
		
		public bool StoragePoolDefineXMLFile(string pathToXMLFile)
		{
			CheckConnected();
			if (!CheckFile(pathToXMLFile))
				return false;
			
			string xml = File.ReadAllText(pathToXMLFile);
			return StoragePoolDefineXMLString(xml);
		}
		
		public bool StoragePoolCreateXMLString(string XMLString)
		{
			CheckConnected();
			IntPtr i = libvirt.virStoragePoolCreateXML(_handle, XMLString.Trim(), 0);
			return (i != IntPtr.Zero);
		}
		
		public bool StoragePoolCreateXMLFile(string pathToXMLFile)
		{
			CheckConnected();
			if (!CheckFile(pathToXMLFile))
				return false;
			
			string xml = File.ReadAllText(pathToXMLFile);
			return StoragePoolCreateXMLString(xml);
		}
		
		// --- INTERNALS --- //
		bool CheckFile(string pathToFile)
		{
			if (!File.Exists(pathToFile.Trim()))
				throw new FileNotFoundException();
			
			return true;
		}
		
		void CheckConnected()
		{
			if (!Connected)
				throw new Exception("Not connected to libvirt daemon. Use \"Connect(string URI)\" first");
		}		
		
		public int MaxNames
		{
			get {return _maxNames;}
			set {_maxNames = value;}
		}
		
	}
	
	/// <summary>
	/// Provides interface for manipulating with domain.
	/// </summary>
	public class slvDomain
	{
		IntPtr _handle;
		slvConnection _conn = null;
		
		// --- CONSTRUCTOR/DESTRUCTOR --- //
		public slvDomain(slvConnection conn, string name)
		{
			_conn = conn;
			CheckConnected();
			ChangeDomain(name);
		}
		
		public slvDomain(slvConnection conn)
		{
			_conn = conn;
			CheckConnected();
		}
		
		~slvDomain()
		{
			if (_handle != IntPtr.Zero)
				libvirt.virDomainFree(_handle);
		}	
		
		// --- ATTRIBUTES --- //
		public bool Loaded
		{
			get {return (_handle != IntPtr.Zero);}
		}
		
		public IntPtr DomainPtr
		{
			get
			{
				return _handle;
			}
		}
		
		/// <summary>
		/// Extract information about a domain. Note that if the connection used to get the domain is limited only a partial 
		/// set of the information can be extracted.
		/// </summary>
		public virDomainInfo Info
		{
			get
			{
				CheckDomain();
				virDomainInfo ret = new virDomainInfo();
				int i = libvirt.virDomainGetInfo(_handle, ret);
				if (i == -1)
					throw new Exception("virDomainGetInfo: -1");
				
				return ret;
			}
		}
				
		/// <summary>
		/// Public name for that domain.
		/// </summary>
		public string Name
		{
			get
			{
				CheckDomain();
				string ret = libvirt.virDomainGetName(_handle);
				if (ret == string.Empty)
					throw new Exception("virDomainGetName: NULL");
				
				return ret;
			}
		}
		
		/// <summary>
		/// Get the type of domain operation system.
		/// </summary>
		public string OSType
		{
			get
			{
				CheckDomain();
				string ret = libvirt.virDomainGetOSType(_handle);
				if (ret == string.Empty)
					throw new Exception("virDomainGetOSType: NULL");
				
				return ret;
			}
		}
		
		/// <summary>
		/// Get the hypervisor ID number for the domain. This only works when the domain is active (running).
		/// </summary>
		public int ID 
		{
			get 
			{
				// Note: No exception here, because this works only that domain is running. 
				// So -1 mean "domain is not running" not an error
				CheckDomain();
				return libvirt.virDomainGetID(_handle);
			}
		}
		
		/// <summary>
		/// Provides a boolean value indicating whether the domain configured to be automatically started when the host machine boots.
		/// Also configure the domain to be automatically started when the host machine boots.
		/// </summary>
		public bool Autostart
		{
			get
			{
				CheckDomain();
				int ret;
				int i = libvirt.virDomainGetAutostart(_handle, out ret);
				if (i == -1)
					throw new Exception("virDomainGetAutostart: -1");
				
				return (ret == 1);
			}
			
			set 
			{
				CheckDomain();
				int i = libvirt.virDomainSetAutostart(_handle, Convert.ToInt32(value));
				if (i == -1)
					throw new Exception("virDomainSetAutostart: -1");
			}
		}
		
		/// <summary>
		/// Provides the maximum number of virtual CPUs supported for the guest VM. If the guest is inactive, 
		/// this is basically the same as virConnectGetMaxVcpus. 
		/// If the guest is running this will reflect the maximum number of virtual CPUs the guest was booted with.
		/// </summary>
		public int MaxVcpus
		{
			get
			{
				CheckDomain();
				int ret = libvirt.virDomainGetMaxVcpus(_handle);
				if (ret == -1)
					throw new Exception("virDomainGetMaxVcpus: -1");
				
				return ret;
			}
		}
		
		/// <summary>
		/// Retrieve the maximum amount of physical memory allocated to a domain. 
		/// Also dynamically change the maximum amount of physical memory allocated to a domain. 
		/// </summary>
		public ulong MaxMemory
		{
			get
			{
				CheckDomain();
				ulong ret = libvirt.virDomainGetMaxMemory(_handle);
				if (ret == 0)
					throw new Exception("virDomainGetMaxMemory: 0");
				return ret;
			}
			
			set
			{
				CheckDomain();
				int i = libvirt.virDomainSetMaxMemory(_handle, value);
				if (i == -1)
					throw new Exception("virDomainSetMaxMemory: -1");	
			}
		}
		
		/// <summary>
		/// Get the UUID for a domain as string. For more information about UUID see RFC4122.
		/// </summary>
		public string UUIDString
		{
			get
			{
				CheckDomain();
				IntPtr ret = Mono.Unix.UnixMarshal.AllocHeap(libvirt.VIR_UUID_STRING_BUFLEN);
				int iRet = libvirt.virDomainGetUUIDString(_handle, ret);
				if (iRet == -1)
					throw new Exception("virDomainGetUUIDString: -1");
				
				return Marshal.PtrToStringAuto(ret);
			}
		}
		
		public char[] UUID
		{
			get
			{
				CheckDomain();
				char[] ret = new char[libvirt.VIR_UUID_BUFLEN];
				int iRet = libvirt.virDomainGetUUID(_handle, ret);
				if (iRet == -1)
					throw new Exception("virDomainGetUUID: -1");
				
				return ret;
			}
		}
		
		/// <summary>
		/// Determine if the domain is currently running.
		/// </summary>
		public bool Active
		{
			get
			{
				CheckDomain();
				int iRet = libvirt.virDomainIsActive(_handle);
				if (iRet == -1)
					throw new Exception("virDomainIsActive: -1");
				
				return (iRet == 1);
			}
		}
		
		/// <summary>
		/// Determine if the domain has a persistent configuration which means it will still exist after shutting down.
		/// </summary>
		public bool Persistent
		{
			get
			{
				CheckDomain();
				int iRet = libvirt.virDomainIsPersistent(_handle);
				if (iRet == -1)
					throw new Exception("virStoragePoolIsPersistent: -1");
				
				return (iRet == 1);
			}
		}
		
		// --- CONTROLING --- //
		
		/// <summary>
		/// Launch a domain. If the call succeed the domain moves from the defined to the running domains pools.
		/// </summary>
		public bool Start()
		{
			CheckDomain();
			return (libvirt.virDomainCreate(_handle) == 0);
		}
		
		/// <summary>
		/// Reboot a domain, the domain object is still usable there after but the domain OS is being stopped for a restart. 
		/// Note that the guest OS may ignore the request.
		/// </summary>
		public bool Reboot()
		{
			CheckDomain();
			return (libvirt.virDomainReboot(_handle, 0) == 0);
		}
		
		/// <summary>
		/// Shutdown a domain, the domain object is still usable there after but the domain OS is being stopped. 
		/// Note that the guest OS may ignore the request. 
		/// </summary>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		public bool Shutdown()
		{
			CheckDomain();
			return (libvirt.virDomainShutdown(_handle) == 0);
		}
				
		/// <summary>
		/// Destroy the domain object. The running instance is shutdown if not down already and all resources used by 
		/// it are given back to the hypervisor. This function may require privileged access.
		/// </summary>
		public bool Destroy()
		{
			CheckDomain();
			return (libvirt.virDomainDestroy(_handle) == 0);
		}
		
		/// <summary>
		/// Suspends an active domain, the process is frozen without further access to CPU resources and I/O,
		///  but the memory used by the domain at the hypervisor level will stay allocated. 
		/// Use Resume() to reactivate the domain. This function may requires privileged access.
		/// </summary>
		/// <returns>
		/// A <see cref="System.Boolean"/>
		/// </returns>
		public bool Suspend()
		{
			CheckDomain();
			return (libvirt.virDomainSuspend(_handle) == 0);
		}
		
		/// <summary>
		/// Resume a suspended domain, the process is restarted from the state where it was frozen by calling Suspend(). 
		/// This function may requires privileged access
		/// </summary>
		public bool Resume()
		{
			CheckDomain();
			return (libvirt.virDomainResume(_handle) == 0);
		}
		
		/// <summary>
		/// This method will suspend a domain and save its memory contents to a file on disk. After the call, 
		/// if successful, the domain is not listed as running anymore (this may be a problem). 
		/// Use nRestore() to restore a domain after saving.
		/// </summary>
		/// <param name="fileName">
		/// A <see cref="System.String"/>path to file name where domain state will be saved.
		/// </param>
		public bool Save(string fileName)
		{
			CheckDomain();
			return (libvirt.virDomainSave(_handle, fileName) == 0);
		}		
		
		/// <summary>
		/// This method will restore a domain saved to disk by Save().
		/// </summary>
		/// <param name="fileName">
		/// A <see cref="System.String"/>path to file with saved domain state.
		/// </param>
		/// <param name="deleteFile">
		/// Delete file after restore. Only if restore was successfull.
		/// </param>
		public bool Restore(string fileName, bool deleteFile)
		{
			CheckDomain();
			if (!File.Exists(fileName))
				throw new FileNotFoundException();
			
			int ret = libvirt.virDomainRestore(_conn.ConnectPtr, fileName);
			if ((ret == 0) && (deleteFile))
				File.Delete(fileName);
			
			return (ret == 0);
		}
		
		/// <summary>
		/// This method will restore a domain saved to disk by Save().
		/// </summary>
		/// <param name="fileName">
		/// A <see cref="System.String"/>path to file with saved domain state.
		/// </param>
		public bool Restore(string fileName)
		{
			return Restore(fileName, false);
		}
	
		// --- DEVICES --- //
		
		/// <summary>
		/// Create a virtual device attachment to backend.
		/// </summary>
		/// <param name="xmlDescription">
		/// XML description of one device. Use empty string for deattach device, i.e. eject CDROM
		/// </param>
		public bool AttachDevice(string xmlDescription)
		{
			CheckDomain();
			return (libvirt.virDomainAttachDevice(_handle, xmlDescription.Trim()) == 0);
		}
		
		/// <summary>
		/// Destroy a virtual device attachment to backend.
		/// </summary>
		/// <param name="xmlDescription">
		/// XML description of one device.
		/// </param>
		public bool DetachDevice(string xmlDescription)
		{
			CheckDomain();
			return (libvirt.virDomainDetachDevice(_handle, xmlDescription.Trim()) == 0);
		}
		
		/// <summary>
		/// This function returns block device (disk) stats for block devices attached to the domain. 
		/// Domains may have more than one block device. To get stats for each you should make multiple calls to this function. 
		/// Individual fields within the stats structure may be returned as -1, 
		/// which indicates that the hypervisor does not support that particular statistic.
		/// </summary>
		/// <param name="path">
		/// The path parameter is the name of the block device. Get this by calling virDomainGetXMLDesc 
		/// and finding the <target dev='...'> attribute within //domain/devices/disk. (For example, "xvda"). 
		/// </param>
		public virDomainBlockStatsStruct BlockStats(string path)
		{
			CheckDomain();
			virDomainBlockStatsStruct ret = new virDomainBlockStatsStruct();
			int iRet = libvirt.virDomainBlockStats(_handle, path.Trim(), ret, Marshal.SizeOf(ret));
			if (iRet == -1)
				throw new Exception("virDomainBlockStats: -1");
			
			return ret;
		}
		
		// --- OTHERS --- //
		
		/// <summary>
		/// This function returns network interface stats for interfaces attached to the domain. 
		/// Domains may have more than one network interface. To get stats for each you should make multiple calls to this function. 
		/// Individual fields within the stats structure may be returned as -1,
		/// which indicates that the hypervisor does not support that particular statistic.
		/// </summary>
		/// <param name="path">
		/// The path parameter is the name of the network interface.
		/// </param>
		public virDomainInterfaceStatsStruct InterfaceStats(string path) // TODO Nefunguje! Dotazat se na IRC, porad vraci nezname jmeno rozhrani
		{
			CheckDomain();
			virDomainInterfaceStatsStruct ret = new virDomainInterfaceStatsStruct();
			int sz = Marshal.SizeOf(ret);
			int iRet = libvirt.virDomainInterfaceStats(_handle, path.Trim(), ret, sz);
			if (iRet == -1)
				throw new Exception("virDomainInterfaceStats: -1");
			
			return ret;
		}		
		
		/// <summary>
		/// This method will dump the core of a domain on a given file for analysis. 
		/// Note that for remote Xen Daemon the file path will be interpreted in the remote host.
		/// </summary>
		/// <param name="fileName">
		/// Path for the core file.
		/// </param>
		public bool CoreDump(string fileName)
		{
			CheckDomain();
			return (libvirt.virDomainCoreDump(_handle, fileName, 0) == 0);
		}
		
		/// <summary>
		/// Provide an XML description of the domain. 
		/// The description may be reused later to relaunch the domain with CreateXML().
		/// </summary>
		/// <param name="flags">
		/// An OR'ed set of <see cref="virDomainXMLFlags"/>
		/// </param>
		public string GetXMLDesc(virDomainXMLFlags flags)
		{
			CheckDomain();
			return libvirt.virDomainGetXMLDesc(_handle, flags);
		}
		
		/// <summary>
		/// Dynamically change the number of virtual CPUs used by the domain. 
		/// Note that this call may fail if the underlying virtualization hypervisor does not support it or 
		/// if growing the number is arbitrary limited. This function requires privileged access to the hypervisor.
		/// </summary>
		/// <param name="nvcpus">
		/// The new number of virtual CPUs for this domain.
		/// </param>
		public bool SetVcpus(uint nvcpus)
		{
			CheckDomain();
			return (libvirt.virDomainSetVcpus(_handle, nvcpus) == 0);
		}
		
		/// <summary>
		/// Undefine a domain but does not stop it if it is running.
		/// </summary>>
		public bool Undefine()
		{
			CheckDomain();
			bool ret = (libvirt.virDomainUndefine(_handle) == 0);
			if (ret)
				_handle = IntPtr.Zero;
			
			return ret;
		}
		
		/// <summary>
		/// Dynamically change the target amount of physical memory allocated to a domain. 
		/// This function may requires privileged access to the hypervisor.
		/// </summary>
		/// <param name="memory">
		/// The memory size in kilobytes
		/// </param>
		public bool SetMemory(ulong memory)
		{
			CheckDomain();
			return (libvirt.virDomainSetMemory(_handle, memory) == 0);
		}
		
		// --- INTERNAL --- //
		
		/// <summary>
		/// Change this object to new domain.
		/// </summary>
		/// <param name="name">
		/// A name of domain.
		/// </param>
		public bool ChangeDomain(string name)
		{
			CheckConnected();
			if (name.Trim() == string.Empty)
				throw new Exception("ChangeDomain: name is empty");
			
			_handle = _conn.DomainLookupByName(name.Trim());
			return (_handle != IntPtr.Zero);
		}
		
		void CheckConnected()
		{
			if (!_conn.Connected)
				throw new Exception("\"conn\" is not connected to libvirt daemon");
		}
		
		void CheckDomain()
		{
			if (_handle == IntPtr.Zero)
				throw new Exception("Domain is not loaded. Use \"ChangeDomain(string name)\" first");
		}		
	}
	
	
	/// <summary>
	/// Interface for handling with storage pool
	/// </summary>
	public class slvStoragePool
	{
		IntPtr _handle = IntPtr.Zero;
		slvConnection _conn = null;
		
		public bool Loaded
		{
			get {return (_handle != IntPtr.Zero);}
		}
		
		void CheckStoragePool()
		{
			if (_handle == IntPtr.Zero)
				throw new Exception("Storage pool is not loaded. Use \"ChangeStoragePool()\" first");
		}
		
		void CheckConnected()
		{
			if (!_conn.Connected)
				throw new Exception("\"conn\" is not connected to libvirt daemon");
		}
		
		public slvStoragePool(slvConnection conn, string name)
		{
			_conn = conn;
			CheckConnected();
			ChangeStoragePool(name);
		}
		
		public slvStoragePool(slvConnection conn)
		{
			_conn = conn;
			CheckConnected();
		}
		
		~slvStoragePool()
		{
			if (_handle != IntPtr.Zero)
				libvirt.virStoragePoolFree(_handle);
		}
		
		/// <summary>
		/// Get volatile information about the storage pool such as free space / usage summary.
		/// </summary>
		public virStoragePoolInfo Info
		{
			get
			{
				CheckStoragePool();
				virStoragePoolInfo spi = new virStoragePoolInfo();
				int iRet = libvirt.virStoragePoolGetInfo(_handle, spi);
				if (iRet == -1)
					throw new Exception("virStoragePoolGetInfo: -1");
				
				return spi;
			}
		}
		
		/// <summary>
		/// Load new information about storage pool to this object.
		/// </summary>
		/// <param name="name">
		/// A name of storage pool.
		/// </param>
		public bool ChangeStoragePool(string name)
		{
			CheckConnected();
			if (name.Trim() == string.Empty)
				throw new Exception("ChangeDomain: name is empty");
			
			_handle = _conn.StoragePoolLookupByName(name.Trim());
			return (_handle != IntPtr.Zero);
		}
		
		public IntPtr PoolPtr
		{
			get {return _handle;}
		}
		
		/// <summary>
		/// Fetch the locally unique name of the storage pool.
		/// </summary>
		public string Name
		{
			get
			{
				CheckStoragePool();
				string ret = libvirt.virStoragePoolGetName(_handle);
				if (ret == string.Empty)
					throw new Exception("virStoragePoolGetName: NULL");
				
				return ret;
			}
		}
		
		/// <summary>
		/// Starts an inactive storage pool.
		/// </summary>
		public bool Start()
		{
			CheckStoragePool();
			return (libvirt.virStoragePoolCreate(_handle, 0) == 0);
		}
		
		/// <summary>
		/// Destroy an active storage pool. This will deactivate the pool on the host, 
		/// but keep any persistent config associated with it. If it has a persistent config it can later be restarted 
		/// with Start().
		/// </summary>
		public bool Stop()
		{
			CheckStoragePool();
			return (libvirt.virStoragePoolDestroy(_handle) == 0);
		}
		
		/// <summary>
		/// Build the underlying storage pool.
		/// </summary>
		public bool Build()
		{
			CheckStoragePool();
			return (libvirt.virStoragePoolBuild(_handle, 0) == 0);
		}
		
		/// <summary>
		/// Request that the pool refresh its list of volumes. This may involve communicating with a remote server, 
		/// and/or initializing new devices at the OS layer.
		public bool Refresh()
		{
			CheckStoragePool();
			return (libvirt.virStoragePoolRefresh(_handle, 0) == 0);
		}
		
		/// <summary>
		/// Fetch an XML document describing all aspects of the storage pool. 
		/// This is suitable for later feeding back into the CreateXML() method.
		/// </summary>
		/// <param name="flags">
		/// A <see cref="virDomainXMLFlags"/>flags for XML format options (set of virDomainXMLFlags)
		/// </param>
		/// <returns>
		/// A a XML document, or String.Empty on error
		/// </returns>
		public string XMLDesc
		{
			get
			{
				CheckStoragePool();
				return libvirt.virStoragePoolGetXMLDesc(_handle, 0);
			}
		}
		
		/// <summary>
		/// Determine if the storage pool is currently running.
		/// </summary>
		public bool Active
		{
			get
			{
				CheckStoragePool();
				int iRet = libvirt.virStoragePoolIsActive(_handle);
				if (iRet == -1)
					throw new Exception("virStoragePoolIsActive: -1");
				
				return (iRet == 1);
			}
		}
		
		/// <summary>
		/// Determine if the storage pool has a persistent configuration which means it will still exist after shutting down.
		/// </summary>
		public bool Persistent
		{
			get
			{
				CheckStoragePool();
				int iRet = libvirt.virStoragePoolIsPersistent(_handle);
				if (iRet == -1)
					throw new Exception("virStoragePoolIsPersistent: -1");
				
				return (iRet == 1);
			}
		}
		
		/// <summary>
		/// Sets the autostart flag or gets the autostart flag.
		/// </summary>
		public bool Autostart
		{
			get
			{
				CheckStoragePool();
				int ret;
				int iRet = libvirt.virStoragePoolGetAutostart(_handle, out ret);
				if (iRet == -1)
					throw new Exception("virStoragePoolGetAutostart: -1");
				
				return (ret == 1);
			}
			
			set
			{
				CheckStoragePool();
				int iRet = libvirt.virStoragePoolSetAutostart(_handle, Convert.ToInt32(value));
				if (iRet == -1)
					throw new Exception("virStoragePoolSetAutostart: -1");
			}
		}
		
		public bool Undefine()
		{
			CheckStoragePool();
			return (libvirt.virStoragePoolUndefine(_handle) == 0);
		}
		
		public string UUIDString
		{
			get
			{
				CheckStoragePool();
				IntPtr buf = Mono.Unix.UnixMarshal.AllocHeap(libvirt.VIR_UUID_STRING_BUFLEN);
				int iRet = libvirt.virStoragePoolGetUUIDString(_handle, buf);
				if (iRet == -1)
					throw new Exception("virStoragePoolGetUUIDString: -1");
				
				return Marshal.PtrToStringAuto(buf);
			}
		}
		
		public int NumOfVolumes
		{
			get
			{
				CheckStoragePool();
				int ret = libvirt.virStoragePoolNumOfVolumes(_handle);
				if (ret == -1)
					throw new Exception("virStoragePoolNumOfVolumes: -1");
				
				return ret;
			}
		}
		
		public string[] ListVolumes
		{
			get
			{
				CheckStoragePool();
				IntPtr ret = Mono.Unix.UnixMarshal.AllocHeap(_conn.MaxNames);
				int iRet = libvirt.virStoragePoolListVolumes(_handle, ret, _conn.MaxNames);
				if (iRet == -1)
					throw new Exception("virStoragePoolListVolumes: -1");
			
				return Mono.Unix.UnixMarshal.PtrToStringArray(iRet, ret);
			}
		}
		
		public IntPtr VolLookupByName(string name)
		{
			CheckStoragePool();
			IntPtr iRet = libvirt.virStorageVolLookupByName(_handle, name.Trim());
			return iRet;
		}
		
		public bool IsVolExists(string name)
		{
			CheckStoragePool();
			return (VolLookupByName(name) != IntPtr.Zero);
		}
		
		public bool VolCreateXMLString(string XMLString)
		{
			CheckStoragePool();
			IntPtr ret = libvirt.virStorageVolCreateXML(_handle, XMLString.Trim(), 0);
			return (ret != IntPtr.Zero);
		}
		
		public bool VolCreateXMLFile(string pathToXMLFile)
		{
			CheckStoragePool();
			if (!File.Exists(pathToXMLFile))
				throw new FileNotFoundException();
			
			string xml = File.ReadAllText(pathToXMLFile);
			return VolCreateXMLString(xml);
		}
		
		public bool VolCreateXMLFrom(string XMLString, IntPtr cloneVol)
		{
			CheckStoragePool();
			IntPtr ret = libvirt.virStorageVolCreateXMLFrom(_handle, XMLString.Trim(), cloneVol, 0);
			return (ret != IntPtr.Zero);
		}
		
		public bool VolCreateXMLFromXMLFile(string pathToXMLFile, IntPtr cloneVol)
		{
			CheckStoragePool();
			if (!File.Exists(pathToXMLFile))
				throw new FileNotFoundException();
			
			string xml = File.ReadAllText(pathToXMLFile);
			return VolCreateXMLFrom(xml, cloneVol);
		}
		
		public char[] UUID
		{
			get
			{
				CheckStoragePool();
				char[] ret = new char[libvirt.VIR_UUID_BUFLEN];
				int iRet = libvirt.virStoragePoolGetUUID(_handle, ret);
				if (iRet == -1)
					throw new Exception("virStoragePoolGetUUID: -1");
				
				return ret;
			}
		}
	}
	
	public class slvStorageVolume 
	{
		IntPtr _handle = IntPtr.Zero;
		slvStoragePool _pool = null;
		
		public slvStorageVolume(slvStoragePool pool, string name)
		{
			_pool = pool;
			CheckStoragePool();
			ChangeVolume(name);
		}
		
		public IntPtr VolPtr
		{
			get {return _handle;}
		}
		
		public bool Loaded
		{
			get {return (_handle != IntPtr.Zero);}
		}
		
		public slvStorageVolume(slvStoragePool pool)
		{
			_pool = pool;
			CheckStoragePool();
		}
		
		~slvStorageVolume()
		{
			if (_handle != IntPtr.Zero)
				libvirt.virStorageVolFree(_handle);
		}
			
		public bool ChangeVolume(string name)
		{
			_handle = libvirt.virStorageVolLookupByName(_pool.PoolPtr, name.Trim());
			return (_handle != IntPtr.Zero);
		}
		
		public virStorageVolInfo Info
		{
			get
			{
				CheckVolume();
				virStorageVolInfo ret = new virStorageVolInfo();
				int iRet = libvirt.virStorageVolGetInfo(_handle, ret);
				if (iRet == -1)
					throw new Exception("virStorageVolGetInfo: -1");
				
				return ret;
			}
		}
		
		public string Name
		{
			get
			{
				CheckVolume();
				string ret = libvirt.virStorageVolGetName(_handle);
				if (ret == string.Empty)
					throw new Exception("virStorageVolGetName: NULL");
				
				return ret;
			}
		}
		
		public string Path
		{
			get
			{
				CheckVolume();
				string ret = libvirt.virStorageVolGetPath(_handle);
				if (ret == string.Empty)
					throw new Exception("virStorageVolGetPath: NULL");
				
				return ret;
			}
		}
		
		// TODO Doesn't work - glibc exception
		public string Key
		{
			get
			{
				CheckVolume();
				string ret = libvirt.virStorageVolGetKey(_handle);
				if (ret == string.Empty)
					throw new Exception("virStorageVolGetKey: NULL");
				
				return ret;
			}
		}
		
		public string XMLDesc
		{
			get
			{
				CheckStoragePool();
				string ret = libvirt.virStorageVolGetXMLDesc(_handle, 0);
				if (ret == string.Empty)
					throw new Exception("virStorageVolGetXMLDesc: NULL");
				
				return ret;
			}
		}
		
		public bool Delete()
		{
			CheckStoragePool();
			bool ret = (libvirt.virStorageVolDelete(_handle, 0) == 0);
			if (ret)
				NotLoaded();
			
			return ret;
		}
		
		void CheckVolume()
		{
			if (_handle == IntPtr.Zero)
				throw new Exception("Storage volume is not loaded. Use \"ChangeVolume()\" first");
		}
		
		void CheckStoragePool()
		{
			if (_pool.PoolPtr == IntPtr.Zero)
				throw new Exception("Storage pool is not loaded. Use \"slvStoragePool.ChangeStoragePool()\" first");
		}
		
		void NotLoaded()
		{
			if (_handle != IntPtr.Zero)
			{
				libvirt.virStorageVolFree(_handle);
				_handle = IntPtr.Zero;
			}
		}
	}
	
	public class slvNetwork
	{
		IntPtr _handle = IntPtr.Zero;
		
		public slvNetwork()
		{
			//
		}
		
		public IntPtr NetwrokPtr
		{
			get {return _handle;}
		}
		
//		void CheckNetwork()
//		{
//			if (_handle == IntPtr.Zero)
//			{
//				throw new Exception(""); // TODO
//			}
//		}
	}
}
